import re
from snappi_ixnetwork.vport import Vport
from collections import namedtuple

class ResourceGroup(object):
    """"""
    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self._store_properties = []
        self._layer1_conf = None
        self.layer1_check = []
        self._layer1_backup = None
        
    def set_group(self):
        self.layer1_check = []
        self._store_properties = []
        self._layer1_conf = self._api.snappi_config.layer1
        if self._layer1_conf is None or len(
                self._layer1_conf ) == 0 or self._is_redundant is True:
            return self.layer1_check
        
        self._cache_properties()
        response = None
        try:
            payload = {"arg1": "/availableHardware", "arg2" : []}
            url = '%s/availableHardware/operations/getChassisWithDetailedResouceGroupsInfo' \
                  % self._api._ixnetwork.href
            response = self._api._request('POST', url, payload)
        except Exception:
            raise Exception("Not able to fetch chassis details. Unable to execute L1 setting")
        self._process_properties(response)
        
        final_arg2 = []
        error_ports = []
        convert_msgs = []
        idle_msgs = []
        ixn_href = self._api._ixnetwork.href
        for property in self._store_properties:
            # We will change those speed from L1 setting
            if property.aggregate is False:
                continue
            if property.group_mode is None:
                error_ports.append(property.port_name)
                continue
            args = [arg['arg1'] for arg in final_arg2]
            url = property.get_url(ixn_href)
            if url is None:
                idle_msgs.append((property.port_name,
                                  property.group_mode))
                continue
            if url not in args:
                convert_msgs.append((property.port_name,
                                     property.group_mode))
                arg2 = {
                    "arg1" : url,
                    "arg2" : property.group_mode
                }
                final_arg2.append(arg2)
        
        if len(error_ports) > 0:
            raise Exception("Please check the speed of these ports ",
                            error_ports)
        if len(idle_msgs) > 0:
            self._api.info("Speed conversion is not require for "
                           "(port.name, speed) : {0}".format(idle_msgs))
        if len(final_arg2) > 0:
            url = "{0}/availableHardware/operations/" \
                  "setresourcegroupsinfo".format(ixn_href)
            payload = {
                "arg1": "/availableHardware",
                "arg2": final_arg2,
                "arg3": True,
                "arg4": True
            }
            try:
                self._api.info("Setting (port.name, speed) : {0}".format(
                            convert_msgs))
                self._api.info(final_arg2)
                self._api._request('POST', url, payload)
            except:
                # todo: redirect to unknown page. Probable IxNetwork issue
                pass
        
        return self.layer1_check

    @property
    def _is_redundant(self):
        is_redundant = False
        if self._layer1_backup is not None:
            if self._layer1_conf == self._layer1_backup:
                self._api.info("Speed change not require "
                               "due to redundant Layer1 config")
                is_redundant = True
                for layer1 in self._layer1_conf:
                    self.layer1_check.append(layer1.name)
        self._layer1_backup = self._layer1_conf
        return is_redundant
        
    def _cache_properties(self):
        port_list = []
        ports = self._api.snappi_config.ports
        for layer1 in self._layer1_conf:
            if layer1.port_names is None or len(
                    layer1.port_names) == 0:
                return
            for port in ports:
                if port in port_list:
                    return
                if port.name in layer1.port_names:
                    location = port.location
                    if location is None:
                        raise Exception("Please configure location to change speed")
                    location_info = self._api.parse_location_info(location)
                    chassis_info = location_info.chassis_info
                    property = StoreProperty(chassis_info,
                                        location_info.card_info,
                                        location_info.port_info,
                                        port.name,
                                        layer1)
                    self._store_properties.append(
                            property)
    
    def get_chassis_card(self, property, response):
        ChassisCardInfo = namedtuple("ChassisCardInfo", ["chassis_id",
                                                         "card_info"])
        chassis_id = 1
        for result in response['result']:
            chassis_dns = result["dns"]
            card_info = None
            if property.chassis_dns == chassis_dns:
                if int(property.card) == 0:
                    card_info = result['cards']
                else:
                    for card in result['cards']:
                        if int(property.card) == int(card['cardId']):
                            card_info = [card]
                if card_info is not None:
                    return ChassisCardInfo(chassis_id,
                                    card_info)
            chassis_id += 1
    
    def _process_properties(self, response):
        for property in self._store_properties:
            info = self.get_chassis_card(property, response)
            if info is None:
                raise Exception("Chassis or card not available for %s"
                                % property.port_name)
            chassis_id = info.chassis_id
            card_info = info.card_info
            for card in card_info:
                if card['cardAggregationMode'] == 'notSupported':
                    continue
                property.aggregate = True
                if self._set_group_mode(property, card,
                        chassis_id) is True:
                    break

    def _set_group_mode(self, property, card, chassis_id):
            for supported_group in card['supportedGroups']:
                group_id = supported_group['id']
                current_mode = supported_group['currentSetting'][
                        'resourceGroupMode']
                for available_setting in supported_group[
                        'availableSettings']:
                    group_mode = available_setting[
                            'resourceGroupMode']
                    for panel_info in available_setting[
                            'panelInfo']:
                        for display_name in panel_info[
                                'activePortsDisplayNames']:
                            if property.port == display_name:
                                l1_name = property.set_property(chassis_id,
                                                      card,
                                                      group_id,
                                                      current_mode,
                                                      group_mode)
                                if l1_name is not None:
                                    if l1_name not in self.layer1_check:
                                        self.layer1_check.append(l1_name)
                                    return True
            return False
            
class StoreProperty(object):
    def __init__(self, chassis, card, port, port_name, layer1):
        self._chassis = chassis
        self._card = card
        self._port = port.lstrip('0')
        self._port_name = port_name
        self._speed = layer1.speed
        self._l1name = layer1.name
        self._chassis_id = None
        self._card_id = None
        self._group_id = None
        self._current_mode = None
        self._group_mode = None
        self._aggregate = False

    @property
    def chassis_dns(self):
        return self._chassis

    @property
    def card(self):
        return self._card
    
    @property
    def port(self):
        return self._port
    
    @property
    def group_mode(self):
        return self._group_mode
    
    @property
    def port_name(self):
        return self._port_name
    
    @property
    def l1name(self):
        return self._l1name
    
    @property
    def aggregate(self):
        return self._aggregate

    @aggregate.setter
    def aggregate(self, value):
        self._aggregate = value
        
    def _get_speed_map(self, card):
        """This check for specific card"""
        if re.search('maz-xb16', card[
                'appId'].lower()) is not None:
            map_speed = 'normal'
        else:
            try:
                map_speed = Vport._SPEED_MODE_MAP[
                        self._speed]
            except Exception:
                raise Exception("Speed %s not available within internal map" %
                                self._speed)
        return map_speed
    
    def set_property(self, chassis_id, card, group_id, current_mode, group_mode):
        map_speed = self._get_speed_map(card)
        if re.search(map_speed, group_mode.lower()) \
                is not None:
            self._chassis_id = chassis_id
            self._card_id = card['cardId']
            self._group_id = group_id
            self._current_mode = current_mode
            self._group_mode = group_mode
            return self._l1name
        return None
    
    def get_url(self, ixn_href):
        if self._current_mode == self._group_mode:
            return None
        url = "{0}/availableHardware/chassis/{1}" \
              "/card/{2}/aggregation/{3}".format(ixn_href,
                             self._chassis_id,
                             self._card_id,
                             self._group_id)
        return url


import time
from snappi_ixnetwork.timer import Timer
from snappi_ixnetwork.customfield import CustomField

class TrafficItem(CustomField):
    """TrafficItem configuration

    Args
    ----
    - ixnetworkapi (Api): instance of the ixnetworkapi class
    
    """
    _RESULT_COLUMNS = [
        ('frames_tx', 'Tx Frames', int),
        ('frames_rx', 'Rx Frames', int),
        ('frames_tx_rate', 'Tx Frame Rate', float),
        ('frames_rx_rate', 'Rx Frame Rate', float),
        ('bytes_tx', 'Tx Bytes', int),
        ('bytes_rx', 'Rx Bytes', int),
        ('loss', 'Loss %', float),
        ('bytes_tx_rate', 'Tx Rate (Bps)', float),
        ('bytes_rx_rate', 'Rx Rate (Bps)', float),
    ]

    _RESULT_LATENCY_STORE_FORWARD = [
        ('minimum_ns', 'Store-Forward Min Latency (ns)', int),
        ('maximum_ns', 'Store-Forward Max Latency (ns)', int),
        ('average_ns', 'Store-Forward Avg Latency (ns)', int),
    ]
    
    _RESULT_LATENCY_CUT_THROUGH = [
        ('minimum_ns', 'Cut-Through Min Latency (ns)', int),
        ('maximum_ns', 'Cut-Through Max Latency (ns)', int),
        ('average_ns', 'Cut-Through Avg Latency (ns)', int),
    ]
    
    _RESULT_TIMESTAMP = [
        ('first_timestamp_ns', 'First TimeStamp', int),
        ('last_timestamp_ns', 'Last TimeStamp', int)
    ]

    _STACK_IGNORE = ['ethernet.fcs', 'pfcPause.fcs']

    _TYPE_TO_HEADER = {
        'ethernet': 'ethernet',
        'pfcPause': 'pfcpause',
        'vlan': 'vlan',
        'ipv4': 'ipv4',
        'ipv6': 'ipv6',
        'tcp': 'tcp',
        'udp': 'udp',
        'gtpu': 'gtpv1',
        'gTPuOptionalFields': 'gtpv1option',
        'custom': 'custom'
    }

    _HEADER_TO_TYPE = {
        'ethernet': 'ethernet',
        'pfcpause': 'pfcPause',
        'vlan': 'vlan',
        'ipv4': 'ipv4',
        'ipv6': 'ipv6',
        'tcp': 'tcp',
        'udp': 'udp',
        'gtpv1': 'gtpu',
        'gtpv1option': 'gTPuOptionalFields',
        'custom': 'custom'
    }

    _BIT_RATE_UNITS_TYPE = {
        'bps': 'bitsPerSec',
        'kbps': 'kbitsPerSec',
        'mbps': 'mbitsPerSec',
        'gbps': 'mbytesPerSec'
    }

    _LATENCY = {
        'cut_through' : 'cutThrough',
        'store_forward' : 'storeForward'
    }

    _PFCPAUSE = {
        'dst': 'pfcPause.header.header.dstAddress',
        'src': 'pfcPause.header.header.srcAddress',
        'ether_type': 'pfcPause.header.header.ethertype',
        'control_op_code': 'pfcPause.header.macControl.controlOpcode',
        'class_enable_vector':
        'pfcPause.header.macControl.priorityEnableVector',
        'pause_class_0': 'pfcPause.header.macControl.pauseQuanta.pfcQueue0',
        'pause_class_1': 'pfcPause.header.macControl.pauseQuanta.pfcQueue1',
        'pause_class_2': 'pfcPause.header.macControl.pauseQuanta.pfcQueue2',
        'pause_class_3': 'pfcPause.header.macControl.pauseQuanta.pfcQueue3',
        'pause_class_4': 'pfcPause.header.macControl.pauseQuanta.pfcQueue4',
        'pause_class_5': 'pfcPause.header.macControl.pauseQuanta.pfcQueue5',
        'pause_class_6': 'pfcPause.header.macControl.pauseQuanta.pfcQueue6',
        'pause_class_7': 'pfcPause.header.macControl.pauseQuanta.pfcQueue7',
    }

    _ETHERNET = {
        'dst': 'ethernet.header.destinationAddress',
        'src': 'ethernet.header.sourceAddress',
        'ether_type': 'ethernet.header.etherType',
        'pfc_queue': 'ethernet.header.pfcQueue',
    }

    _VLAN = {
        'id': 'vlan.header.vlanTag.vlanID',
        'cfi': 'vlan.header.vlanTag.cfi',
        'priority': 'vlan.header.vlanTag.vlanUserPriority',
        'protocol': 'vlan.header.protocolID'
    }

    _IPV4 = {
        'version': 'ipv4.header.version',
        'header_length': 'ipv4.header.headerLength',
        'priority': '_ipv4_priority',
        'total_length': 'ipv4.header.totalLength',
        'identification': 'ipv4.header.identification',
        'reserved': 'ipv4.header.flags.reserved',
        'dont_fragment': 'ipv4.header.flags.fragment',
        'more_fragments': 'ipv4.header.flags.lastFragment',
        'fragment_offset': 'ipv4.header.fragmentOffset',
        'time_to_live': 'ipv4.header.ttl',
        'protocol': 'ipv4.header.protocol',
        'header_checksum': 'ipv4.header.checksum',
        'src': 'ipv4.header.srcIp',
        'dst': 'ipv4.header.dstIp',
    }

    _IPV6 = {
        'version' : 'ipv6.header.versionTrafficClassFlowLabel.version',
        'traffic_class' : 'ipv6.header.versionTrafficClassFlowLabel.trafficClass',
        'flow_label' : 'ipv6.header.versionTrafficClassFlowLabel.flowLabel',
        'payload_length' : 'ipv6.header.payloadLength',
        'next_header' : 'ipv6.header.nextHeader',
        'hop_limit' : 'ipv6.header.hopLimit',
        'src' : 'ipv6.header.srcIP',
        'dst' : 'ipv6.header.dstIP'
    }
    
    _TOS = {
        "precedence": "ipv4.header.priority.tos.precedence",
        "delay": "ipv4.header.priority.tos.delay",
        "throughput": "ipv4.header.priority.tos.throughput",
        "reliability": "ipv4.header.priority.tos.reliability",
        "monetary": "ipv4.header.priority.tos.monetary",
        "unused": "ipv4.header.priority.tos.unused"
    }

    _TCP = {
        "src_port": "tcp.header.srcPort",
        "dst_port": "tcp.header.dstPort",
        "ecn_ns": "tcp.header.ecn.nsBit",
        "ecn_cwr": "tcp.header.ecn.cwrBit",
        "ecn_echo": "tcp.header.ecn.ecnEchoBit",
        "ctl_urg": "tcp.header.controlBits.urgBit",
        "ctl_ack": "tcp.header.controlBits.ackBit",
        "ctl_psh": "tcp.header.controlBits.pshBit",
        "ctl_rst": "tcp.header.controlBits.rstBit",
        "ctl_syn": "tcp.header.controlBits.synBit",
        "ctl_fin": "tcp.header.controlBits.finBit",
    }

    _UDP = {
        "src_port": "udp.header.srcPort",
        "dst_port": "udp.header.dstPort",
        "length": "udp.header.length",
        "checksum": "udp.header.checksum",
    }
    
    _GTPV1 = {
        "version" : "gtpu.header.version",
        "protocol_type" : "gtpu.header.pt",
        "reserved" : "gtpu.header.reserved",
        "e_flag" : "gtpu.header.e",
        "s_flag" : "gtpu.header.s",
        "pn_flag" : "gtpu.header.n",
        "message_type" : "tpu.header.type",
        "message_length" : "gtpu.header.totalLength",
        "teid" : "gtpu.header.teid"
    }

    _GTPV1OPTION = {
        "squence_number" : "gTPuOptionalFields.header.sequenceNumber",
        "n_pdu_number" : "gTPuOptionalFields.header.npduNumber",
        "next_extension_header_type" : "gTPuOptionalFields.header.nextExtHdrField"
    }

    _CUSTOM = '_custom_headers'

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self.has_latency = False
        self._flow_timeout = 10

    def config(self):
        """Configure config.flows onto Ixnetwork.Traffic.TrafficItem
        
        CRUD
        ----
        - DELETE any TrafficItem.Name that does not exist in config.flows
        - CREATE TrafficItem for any config.flows[*].name that does not exist
        - UPDATE TrafficItem for any config.flows[*].name that exists
        """
        ixn_traffic_item = self._api._traffic_item
        self.flows_has_latency = []
        self.flows_has_timestamp = []
        self.flows_has_loss = []
        self.latency_mode = None
        self._api._remove(ixn_traffic_item, self._api.snappi_config.flows)
        if len(self._api.snappi_config.flows) > 0:
            for flow in self._api.snappi_config.flows:
                self._endpoint_validation(flow)
                args = {
                    'Name': flow.name,
                    'TrafficItemType': 'l2L3',
                    'TrafficType': self._get_traffic_type(flow),
                    'SrcDestMesh': self._get_mesh_type(flow)
                }
                ixn_traffic_item.find(Name='^%s$' % self._api.special_char(flow.name),
                                      TrafficType=args['TrafficType'])
                if len(ixn_traffic_item) == 0:
                    ixn_traffic_item.add(**args)
                else:
                    self._update(ixn_traffic_item, **args)
                self._configure_endpoint(ixn_traffic_item.EndpointSet,
                                         flow.tx_rx)
                metrics = flow._properties.get('metrics')
                if metrics is not None \
                        and metrics.enable is True:
                    self._configure_tracking(flow, ixn_traffic_item.Tracking)
                    latency = metrics._properties.get('latency')
                    if latency is not None and latency.enable is True:
                        self.flows_has_latency.append(flow.name)
                        self._process_latency(latency)
                    timestamps = metrics._properties.get('timestamps')
                    if timestamps is True:
                        self.flows_has_timestamp.append(flow.name)
                    loss = metrics._properties.get('loss')
                    if loss is True:
                        self.flows_has_loss.append(flow.name)
                ixn_ce = ixn_traffic_item.ConfigElement.find()
                hl_stream_count = len(ixn_traffic_item.HighLevelStream.find())
                self._configure_stack(ixn_ce, flow.packet)
                self._configure_size(ixn_ce, flow.size)
                self._configure_rate(ixn_ce, flow.rate)
                self._configure_tx_control(ixn_ce, hl_stream_count, flow.duration)
            self._configure_options()
            self._configure_latency()

    def _process_latency(self, latency):
        if self.latency_mode is None:
            if latency.mode is not None:
                self.latency_mode = latency.mode
            else:
                self.latency_mode = 'store_forward'
        else:
            if latency.mode is not None and \
                    self.latency_mode != latency.mode:
                raise Exception("Latency mode needs to be same for all flows")

    def _configure_latency(self):
        ixn_latency = self._api._traffic.Statistics.Latency
        if self.latency_mode is not None:
            self.has_latency = True
            ixn_CpdpConvergence = self._api._traffic.Statistics.CpdpConvergence
            ixn_CpdpConvergence.Enabled = False
            ixn_latency.Enabled = True
            ixn_latency.Mode = TrafficItem._LATENCY[
                self.latency_mode]
        else:
            self.has_latency = False
            ixn_latency.Enabled = False
    
    def _configure_tracking(self, flow, ixn_tracking):
        """Set tracking options"""
        ixn_tracking.find()
        tracking_options = ['trackingenabled0']
        if flow.tx_rx.choice == 'device':
            tracking_options.append('sourceDestPortPair0')
        if set(tracking_options) != set(ixn_tracking.TrackBy):
            ixn_tracking.update(TrackBy=tracking_options)

    def _configure_options(self):
        enable_min_frame_size = False
        for flow in self._api.snappi_config.flows:
            if (len(flow.packet) == 1 and flow.packet[
                    0].parent.choice == 'pfcpause'):
                enable_min_frame_size = True
                break
        if self._api._traffic.EnableMinFrameSize != enable_min_frame_size:
            self._api._traffic.EnableMinFrameSize = enable_min_frame_size

    def _endpoint_validation(self, flow):
        if flow.tx_rx.choice is None:
            raise ValueError('%s Flow.tx_rx property cannot be None' %
                             flow.name)
        if flow.tx_rx.choice == 'device':
            device = flow.tx_rx.device
            if not isinstance(device.tx_names, list) or \
                    not isinstance(device.rx_names, list):
                raise ValueError("device tx_names and rx_names must be a list "
                                 "in flow %s" % flow.name)
            if len(device.tx_names) != len(set(device.tx_names)):
                raise ValueError("All names in device tx_names "
                                 "must be unique for flow %s" % flow.name)
            if len(device.rx_names) != len(set(device.rx_names)):
                raise ValueError("All names in device rx_names "
                                 "must be unique for flow %s" % flow.name)
    
    def _get_mesh_type(self, flow):
        if flow.tx_rx.choice == 'port':
            mesh_type = 'oneToOne'
        else:
            device = flow.tx_rx.device
            if device.mode == 'mesh' or \
                    device.mode is None:
                mesh_type = 'manyToMany'
            else:
                mesh_type = 'oneToOne'
                if len(device.tx_names) != len(device.rx_names):
                    raise ValueError("Length of device tx_names and rx_names "
                             "must be same for device mode ONE_TO_ONE in flow %s" % flow.name)
        return mesh_type
    
    def _get_traffic_type(self, flow):
        if flow.tx_rx.choice == 'port':
            encap = 'raw'
        else:
            encap_list = []
            device = flow.tx_rx.device
            if device.tx_names is not None:
                for tx_name in device.tx_names:
                    encap_list.append(self._api.get_device_encap(
                            tx_name))
            if device.rx_names is not None:
                for rx_name in device.rx_names:
                    encap_list.append(self._api.get_device_encap(
                            rx_name))
            if len(set(encap_list)) == 1:
                encap = encap_list[0]
            else:
                raise Exception("All devices identified in tx_names and rx_names "
                                "must be of same type in flow %s" %flow.name)
        return encap

    def _configure_endpoint(self, ixn_endpoint_set, endpoint):
        """Transform flow.tx_rx to /trafficItem/endpointSet
        The model allows for only one endpointSet per traffic item
        """
        args = {'Sources': [], 'Destinations': []}
        if (endpoint.choice == "port"):
            args['Sources'].append(
                self._api.get_ixn_object(
                    endpoint.port.tx_name).Protocols.find().href)
            if endpoint.port.rx_name != None:
                args['Destinations'].append(
                    self._api.get_ixn_object(
                        endpoint.port.rx_name).Protocols.find().href)
        else:
            for device_name in endpoint.device.tx_names:
                args['Sources'].append(self._api.get_ixn_href(device_name))
            for device_name in endpoint.device.rx_names:
                args['Destinations'].append(self._api.get_ixn_href(device_name))
        ixn_endpoint_set.find()
        if len(ixn_endpoint_set) > 1:
            ixn_endpoint_set.remove()
        if len(ixn_endpoint_set) == 0:
            ixn_endpoint_set.add(**args)
        elif ixn_endpoint_set.Sources != args[
                'Sources'] or ixn_endpoint_set.Destinations != args[
                    'Destinations'] or len(ixn_endpoint_set.parent.ConfigElement.find()) == 0:
            self._update(ixn_endpoint_set, **args)

    def _update(self, ixn_object, **kwargs):
        from ixnetwork_restpy.base import Base
        update = False
        for name, value in kwargs.items():
            if isinstance(value, list) and len(value) > 0 and isinstance(
                    value[0], Base):
                value = [item.href for item in value]
            elif isinstance(value, Base):
                value = value.href
            if getattr(ixn_object, name) != value:
                update = True
        if update is True:
            ixn_object.update(**kwargs)

    def _configure_stack(self, ixn_stream, headers):
        """Transform flow.packets[0..n] to /traffic/trafficItem/configElement/stack
        The len of the headers list is the definitive list which means add/remove
        any stack items so that the stack list matches the headers list.
        If the headers list is empty then use the traffic generator default stack.
        """
        headers = self.adjust_header(headers)
        ixn_stack = ixn_stream.Stack.find()
        for i in range(0, len(headers)):
            header = headers[i]
            if len(ixn_stack) <= i:
                stack = self._add_stack(ixn_stream, stack, header)
            else:
                stack_type_id = ixn_stack[i].StackTypeId
                if stack_type_id in self._STACK_IGNORE:
                    stack = self._add_stack(ixn_stream, stack, header)
                elif stack_type_id not in TrafficItem._TYPE_TO_HEADER:
                    stack = self._add_stack(ixn_stream, ixn_stack[i], header)
                elif TrafficItem._TYPE_TO_HEADER[
                        stack_type_id] != header.choice:
                    stack = self._add_stack(ixn_stream, ixn_stack[i], header)
                else:
                    stack = ixn_stack[i]
            self._configure_field(stack.Field, header)

        # scan and compare new stack to overcome IxNetwork stack serialization
        # then remove additional stack
        if len(headers) == 0:
            return
        stacks_to_remove = []
        ixn_stack = ixn_stream.Stack.find()
        header_index = 0
        for i in range(0, len(ixn_stack)):
            stack_type_id = ixn_stack[i].StackTypeId
            if stack_type_id in self._STACK_IGNORE:
                continue
            if len(headers) <= header_index:
                stacks_to_remove.append(ixn_stack[i])
                continue
            if TrafficItem._TYPE_TO_HEADER[
                    stack_type_id] != headers[header_index].choice:
                stacks_to_remove.append(ixn_stack[i])
            else:
                header_index += 1
        for stack in stacks_to_remove[::-1]:
            stack.Remove()

    def _add_stack(self, ixn_stream, ixn_stack, header):
        type_id = '^%s$' % TrafficItem._HEADER_TO_TYPE[header.choice]
        template = self._api._traffic.ProtocolTemplate.find(
            StackTypeId=type_id)
        stack_href = ixn_stack.AppendProtocol(template)
        return ixn_stream.Stack.read(stack_href)

    def _configure_field(self, ixn_field, header, field_choice=False):
        """Transform flow.packets[0..n].header.choice to /traffic/trafficItem/configElement/stack/field
        """
        field_map = getattr(self, '_%s' % header.choice.upper())
        packet = getattr(header, header.choice)
        if isinstance(field_map, dict) is False:
            method = getattr(self, field_map)
            method(ixn_field, packet)
            return

        for packet_field_name in dir(packet):
            if packet_field_name in field_map:
                pattern = getattr(packet, packet_field_name)
                field_type_id = field_map[packet_field_name]
                self._configure_pattern(ixn_field, field_type_id, pattern,
                                        field_choice)

    def _configure_pattern(self,
                           ixn_field,
                           field_type_id,
                           pattern,
                           field_choice=False):
        custom_field = getattr(self, field_type_id, None)
        if custom_field is not None:
            if pattern.choice is not None:
                custom_field(ixn_field, pattern)
            return
    
        ixn_field = ixn_field.find(FieldTypeId=field_type_id)
        self._set_default(ixn_field, field_choice)
    
        if pattern.choice is None:
            return
    
        if pattern.choice == 'value':
            ixn_field.update(Auto=False,
                             ActiveFieldChoice=field_choice,
                             ValueType='singleValue',
                             SingleValue=pattern.value)
        elif pattern.choice == 'values':
            ixn_field.update(Auto=False,
                             ActiveFieldChoice=field_choice,
                             ValueType='valueList',
                             ValueList=pattern.values)
        elif pattern.choice == 'increment':
            ixn_field.update(Auto=False,
                             ValueType='increment',
                             ActiveFieldChoice=field_choice,
                             StartValue=pattern.increment.start,
                             StepValue=pattern.increment.step,
                             CountValue=pattern.increment.count)
        elif pattern.choice == 'decrement':
            ixn_field.update(Auto=False,
                             ValueType='decrement',
                             ActiveFieldChoice=field_choice,
                             StartValue=pattern.decrement.start,
                             StepValue=pattern.decrement.step,
                             CountValue=pattern.decrement.count)
        elif pattern.choice == 'random':
            ixn_field.update(Auto=False,
                             ActiveFieldChoice=field_choice,
                             ValueType='repeatableRandomRange',
                             MinValue=pattern.random.min,
                             MaxValue=pattern.random.max,
                             StepValue=pattern.random.step,
                             Seed=pattern.random.seed,
                             CountValue=pattern.random.count)
        else:
            # TBD: add to set_config errors - invalid pattern specified
            pass
    
        if pattern.metric_group is not None:
            ixn_field.TrackingEnabled = True
            self._api.ixn_objects[pattern.metric_group] = ixn_field.href

    def _set_default(self, ixn_field, field_choice):
        """We are setting all the field to default. Otherwise test is keeping the same value from previous run."""
        if ixn_field.ReadOnly:
            return
        
        if ixn_field.SupportsAuto:
            if ixn_field.Auto is not True:
                ixn_field.Auto = True
        else:
            ixn_field.update(Auto=False,
                             ActiveFieldChoice=field_choice,
                             ValueType='singleValue',
                             SingleValue=ixn_field.DefaultValue)

    def _configure_size(self, ixn_stream, size):
        """ Transform frameSize flows.size to /traffic/trafficItem[*]/configElement[*]/frameSize
        """
        if size.choice is None:
            return
        ixn_frame_size = ixn_stream.FrameSize
        args = {}
        if size.choice == 'fixed':
            args['Type'] = "fixed"
            args['FixedSize'] = size.fixed
        elif size.choice == 'increment':
            args['Type'] = "increment"
            args['IncrementFrom'] = size.increment.start
            args['IncrementTo'] = size.increment.end
            args['IncrementStep'] = size.increment.step
        elif size.choice == 'random':
            args['Type'] = "random"
            args['RandomMin'] = size.random.min
            args['RandomMax'] = size.random.max
        else:
            print('Warning - We need to implement this %s choice' %
                  size.choice)
        self._update(ixn_frame_size, **args)

    def _configure_rate(self, ixn_stream, rate):
        """ Transform frameRate flows.rate to /traffic/trafficItem[*]/configElement[*]/frameRate
        """
        if rate.choice is None:
            return
        ixn_frame_rate = ixn_stream.FrameRate
        args = {}
        value = None
        if rate.choice == 'percentage':
            args['Type'] = 'percentLineRate'
            value = rate.percentage
        elif rate.choice == 'pps':
            args['Type'] = 'framesPerSecond'
            value = rate.pps
        else:
            args['Type'] = 'bitsPerSecond'
            args['BitRateUnitsType'] = TrafficItem._BIT_RATE_UNITS_TYPE[
                rate.choice]
            value = getattr(rate, rate.choice)
        args['Rate'] = value
        self._update(ixn_frame_rate, **args)

    def _configure_delay(self, delay, args):
        if delay.choice is not None:
            value = getattr(delay, delay.choice, None)
            if value is None:
                raise Exception("Delay must be of type <int>")
            if isinstance(value, float):
                self._api.warning("Cast Delay to <int> due to software limitation")
            if delay.choice == 'microseconds':
                args['StartDelayUnits'] = 'nanoseconds'
                args['StartDelay'] = value * 1000
            else:
                args['StartDelayUnits'] = delay.choice
                args['StartDelay'] = value
    
    def _configure_tx_control(self, ixn_stream, hl_stream_count, duration):
        """Transform duration flows.duration to /traffic/trafficItem[*]/configElement[*]/TransmissionControl
        """
        if duration.choice is None:
            return
        ixn_tx_control = ixn_stream.TransmissionControl
        args = {}
        if duration.choice == 'continuous':
            args['Type'] = 'continuous'
            args['MinGapBytes'] = duration.continuous.gap
            self._configure_delay(duration.continuous.delay, args)
        elif duration.choice == 'fixed_packets':
            args['Type'] = 'fixedFrameCount'
            args['FrameCount'] = duration.fixed_packets.packets / hl_stream_count
            args['MinGapBytes'] = duration.fixed_packets.gap
            self._configure_delay(duration.fixed_packets.delay, args)
        elif duration.choice == 'fixed_seconds':
            args['Type'] = 'fixedDuration'
            args['Duration'] = duration.fixed_seconds.seconds
            args['MinGapBytes'] = duration.fixed_seconds.gap
            self._configure_delay(duration.fixed_seconds.delay, args)
        elif duration.choice == 'burst':
            if duration.burst.bursts is not None \
                    and int(duration.burst.bursts) > 0:
                args['Type'] = 'burstFixedDuration'
                args['RepeatBurst'] = duration.burst.bursts
            else:
                args['Type'] = 'custom'
            args['BurstPacketCount'] = duration.burst.packets
            args['MinGapBytes'] = duration.burst.gap
            args['EnableInterBurstGap'] = True
            inter_burst_gap = duration.burst.inter_burst_gap
            if inter_burst_gap.choice is not None:
                value = getattr(inter_burst_gap, inter_burst_gap.choice, None)
                if value is None:
                    raise Exception("Inter packet gap mush be of type some <int> value")
                if inter_burst_gap.choice == 'microseconds':
                    args['InterBurstGap'] = value * 1000
                    args['InterBurstGapUnits'] = 'nanoseconds'
                else:
                    args['InterBurstGap'] = value
                    args['InterBurstGapUnits'] = inter_burst_gap.choice
        self._update(ixn_tx_control, **args)

    def transmit(self, request):
        """Set flow transmit
        1) If start then start any device protocols that are traffic dependent
        2) If start then generate and apply traffic
        3) Execute requested transmit action (start|stop|pause|resume)
        """
        regex = ''
        flow_names = [flow.name for flow in self._api._config.flows]
        if request and request.flow_names:
            flow_names = request.flow_names
        if len(flow_names) == 1:
            regex = '^%s$' % self._api.special_char(flow_names)[0]
        elif len(flow_names) > 1:
            regex = '^(%s)$' % '|'.join(
                self._api.special_char(flow_names)
            )

        if request.state == 'start':
            if len(self._api._topology.find()) > 0:
                with Timer(self._api, 'Devices start'):
                    self._api._ixnetwork.StartAllProtocols('sync')
                    self._api.check_protocol_statistics()
            if len(self._api._traffic_item.find()) == 0:
                return
            self._api._traffic_item.find(State='^unapplied$')
            if len(self._api._traffic_item) > 0:
                with Timer(self._api, 'Flows generate/apply'):
                    self._api._traffic_item.Generate()
                    self._api._traffic.Apply()
            self._api._traffic_item.find(State='^started$')
            if len(self._api._traffic_item) == 0:
                with Timer(self._api, 'Flows clear statistics'):
                    self._api._ixnetwork.ClearStats(
                        ['waitForPortStatsRefresh', 'waitForTrafficStatsRefresh'])
            self._api.capture._start_capture()
        self._api._traffic_item.find(Name=regex)
        if len(self._api._traffic_item) > 0:
            if request.state == 'start':
                self._api._traffic_item.find(Name=regex, Suspend=True,
                                             State='^started$')
                if len(self._api._traffic_item) > 0:
                    with Timer(self._api, 'Flows resume'):
                        self._api._traffic_item.PauseStatelessTraffic(False)
                self._api._traffic_item.find(Name=regex, State='^stopped$')
                if len(self._api._traffic_item) > 0:
                    with Timer(self._api, 'Flows start'):
                        self._api._traffic_item.StartStatelessTrafficBlocking()
            elif request.state == 'stop':
                self._api._traffic_item.find(Name=regex, State='^started$')
                if len(self._api._traffic_item) > 0:
                    with Timer(self._api, 'Flows stop'):
                        self._api._traffic_item.StopStatelessTrafficBlocking()
            elif request.state == 'pause':
                self._api._traffic_item.find(Name=regex, State='^started$')
                if len(self._api._traffic_item) > 0:
                    with Timer(self._api, 'Flows pause'):
                        self._api._traffic_item.PauseStatelessTraffic(True)
        if request.state == 'stop':
            if len(self._api._topology.find()) > 0:
                with Timer(self._api, 'Devices stop'):
                    self._api._ixnetwork.StopAllProtocols('sync')

    def _set_result_value(self,
                          row,
                          column_name,
                          column_value,
                          column_type=str):
        if len(self._column_names
               ) > 0 and column_name not in self._column_names:
            return
        try:
            row[column_name] = column_type(column_value)
        except:
            if column_type.__name__ in ['float', 'int']:
                row[column_name] = 0
            else:
                row[column_type] = column_value

    def _get_state(self, state):
        """IxNetwork traffic item states
            error, locked, started, 
            startedWaitingForStats, startedWaitingForStreams, stopped, 
            stoppedWaitingForStats, txStopWatchExpected, unapplied
        """
        started_states = [
            'txStopWatchExpected', 'locked', 'started',
            'startedWaitingForStats', 'startedWaitingForStreams',
            'stoppedWaitingForStats'
        ]
        if state in started_states:
            return 'started'
        else:
            return 'stopped'
    
    def results(self, request):
        """Return flow results
        """
        # setup parameters
        self._column_names = request._properties.get('metric_names')
        if self._column_names is None:
            self._column_names = []
        elif not isinstance(self._column_names, list):
            msg = "Invalid format of column_names passed {},\
                    expected list".format(self._column_names)
            raise Exception(msg)

        flow_names = request._properties.get('flow_names')
        has_request_flow = True
        if flow_names is None or len(flow_names) == 0:
            has_request_flow = False
            flow_names = [flow.name for flow in self._api._config.flows]
        elif not isinstance(flow_names, list):
            msg = "Invalid format of flow_names passed {},\
                    expected list".format(flow_names)
            raise Exception(msg)
        final_flow_names = []
        for flow in self._api._config.flows:
            metrics = flow._properties.get('metrics')
            if metrics is None:
                continue
            if metrics.enable is True \
                    and flow.name in flow_names:
                final_flow_names.append(flow.name)
        if len(final_flow_names) == 0:
            msg = """
            To fetch flow metrics at least one flow shall have metric enabled
            """
            raise Exception(msg.strip())
        diff = set(flow_names).difference(final_flow_names)
        if len(diff) > 0 and has_request_flow is True:
            raise Exception("Flow metrics is not enabled on flows {}".format(
                    "".join(list(diff))))
        flow_names = final_flow_names
        regfilter = {'property': 'name', 'regex': '.*'}
        regfilter['regex'] = '^(%s)$' % '|'.join(
            self._api.special_char(flow_names)
        )

        flow_count = len(flow_names)
        ixn_page = self._api._ixnetwork.Statistics.View.find(Caption="Flow Statistics").Page
        if ixn_page.PageSize < flow_count:
            ixn_page.PageSize = flow_count

        # initialize result values
        flow_rows = {}
        for traffic_item in self._api.select_traffic_items(
                traffic_item_filters=[regfilter]).values():
            for stream in traffic_item['highLevelStream']:
                for rx_port_name in stream['rxPortNames']:
                    flow_row = {}
                    self._set_result_value(flow_row, 'name', traffic_item['name'])
                    self._set_result_value(flow_row, 'transmit', self._get_state(traffic_item['state']))
                    self._set_result_value(flow_row, 'port_tx', stream['txPortName'])
                    self._set_result_value(flow_row, 'port_rx', rx_port_name)
                    # init all columns with corresponding zero-values so that
                    # the underlying dictionary contains all requested columns
                    # in an event of unwanted exceptions
                    for external_name, _, external_type in self._RESULT_COLUMNS:
                        self._set_result_value(
                            flow_row, external_name, 0, external_type
                        )
                    flow_rows[traffic_item['name'] + stream['txPortName'] + rx_port_name] = flow_row
                    
        flow_stat = self._api.assistant.StatViewAssistant(
            'Flow Statistics')
        for row in flow_stat.Rows:
            name = row['Traffic Item']
            if len(flow_names) > 0 and name not in flow_names:
                continue
            tx_port = row['Tx Port']
            rx_port = row['Rx Port']
            if name + tx_port + rx_port in flow_rows:
                flow_row = flow_rows[name + tx_port + rx_port]
                if float(row['Tx Frame Rate']) > 0 or int(row['Tx Frames']) == 0:
                    flow_row['transmit'] = 'started'
                else:
                    flow_row['transmit'] = 'stopped'
                for external_name, internal_name, external_type in self._RESULT_COLUMNS:
                    # keep plugging values for next columns even if the
                    # current one raises exception
                    try:
                        self._set_result_value(
                            flow_row, external_name, row[internal_name],
                            external_type
                        )
                    except Exception:
                        # TODO print a warning maybe ?
                        pass
                if name in self.flows_has_latency:
                    self._construct_latency(flow_row, row)
                if name in self.flows_has_timestamp:
                    self._construct_timestamp(flow_row, row)
                if name not in self.flows_has_loss:
                    flow_row.pop('loss')
        return list(flow_rows.values())

    def _construct_latency(self, flow_row, row):
        if self.latency_mode == 'store_forward':
            latency_map = TrafficItem._RESULT_LATENCY_STORE_FORWARD
        else:
            latency_map = TrafficItem._RESULT_LATENCY_CUT_THROUGH
        latency_result = {}
        for external_name, internal_name, external_type in latency_map:
            if internal_name not in row.Columns:
                raise Exception("Could not fetch column %s in latency metrics" %internal_name)
            try:
                self._set_result_value(latency_result, external_name,
                                       row[internal_name], external_type)
            except Exception as exception_err:
                self._api.warning("Could not fetch latency metrics: %s" % exception_err)
                pass
        if len(latency_result) > 0:
            flow_row['latency'] = latency_result
    
    def _construct_timestamp(self, flow_row, row):
        timestamp_result = {}
        for external_name, internal_name, external_type in TrafficItem._RESULT_TIMESTAMP:
            if internal_name not in row.Columns:
                raise Exception("Could not fetch column %s in timestamp metrics" %internal_name)
            
            try:
                val = row[internal_name].split(':')
                mul = [3600, 60, 1]
                sv = sum([
                    int(float(v) * 10 ** 9 + 0.1) * mul[i]
                    for i, v in enumerate(val)
                ])
                self._set_result_value(timestamp_result, external_name,
                                       sv, external_type)
            except Exception as exception_err:
                self._api.warning("Could not fetch timestamp metrics: %s" % exception_err)
                pass
        if len(timestamp_result) > 0:
            flow_row['timestamps'] = timestamp_result


import re
from collections import namedtuple

class ConfigureBgp(object):
    _BGP_AS_SET_MODE = {
        'do_not_include_as': 'dontincludelocalas',
        'include_as_seq': 'includelocalasasasseq',
        'include_as_set': 'includelocalasasasset',
        'include_as_seq_confed': 'includelocalasasasseqconfederation',
        'include_as_set_confed': 'includelocalasasassetconfederation',
        'prepend_as_to_first_segment': 'prependlocalastofirstsegment'
    }
    
    _BGP_AS_MODE = {
        'do_not_include_local_as': 'dontincludelocalas',
        'include_as_seq': 'includelocalasasasseq',
        'include_as_set': 'includelocalasasasset',
        'include_as_confed_seq': 'includelocalasasasseqconfederation',
        'include_as_confed_set': 'includelocalasasassetconfederation',
        'prepend_to_first_segment': 'prependlocalastofirstsegment'
    }
    
    _BGP_SEG_TYPE = {
        'as_seq': 'asseq',
        'as_set': 'asset',
        'as_confed_seq': 'asseqconfederation',
        'as_confed_set': 'assetconfederation'
    }
    
    _BGP_COMMUNITY_TYPE = {
        'manual_as_number': 'manual',
        'no_export': 'noexport',
        'no_advertised': 'noadvertised',
        'no_export_subconfed': 'noexport_subconfed',
        'llgr_stale': 'llgr_stale',
        'no_llgr': 'no_llgr'
    }
    
    _BGP_SR_TE = {
        'policy_type' : {
            'ixn_attr' : 'policyType',
            'default' : 'ipv4'
        },
        'distinguisher' : {
            'ixn_attr' : 'distinguisher',
            'default' : '1'
        },
        'color' : {
            'ixn_attr' : 'policyColor',
            'default' : '101'
        },
        'ipv4_endpoint' : {
            'ixn_attr' : 'endPointV4',
            'default' : '0.0.0.0'
        },
        'ipv6_endpoint' : {
            'ixn_attr' : 'endPointV6',
            'default' : '0:0:0:0:0:0:0:0'
        }
    }

    _SRTE_NEXT_HOP = {
        'next_hop_mode' : {
            'ixn_attr': 'setNextHop',
            'default': 'sameaslocalip',
            'enum_map': {
                'local_ip' : 'sameaslocalip',
                'manual' : 'manually'
            }
        },
        'next_hop_address_type' : {
            'ixn_attr' : 'setNextHopIpType',
            'default' : 'ipv4'
        },
        'ipv4_address' : {
            'ixn_attr' : 'ipv4NextHop',
            'default' : '0.0.0.0'
        },
        'ipv6_address' : {
            'ixn_attr' : 'ipv6NextHop',
            'default' : '::'
        }
    }
    
    _SRTE_ADDPATH = {
        'path_id' : {
            'ixn_attr' : 'addPathId',
            'default' : '1'
        }
    }

    _SRTE_AS_PATH = {
        'override_peer_as_set_mode' : {
            'ixn_attr' : 'overridePeerAsSetMode',
            'default' : False
        },
        'as_set_mode' : {
            'ixn_attr' : 'asSetMode',
            'default' : 'includelocalasasasseq',
            'enum_map' : _BGP_AS_MODE
        }
    }
    
    _SRTE_ASPATH_SEGMENT = {
        'segment_type' : {
            'ixn_attr' : 'segmentType',
            'default' : 'asset',
            'enum_map' : _BGP_SEG_TYPE
        }
    }
    
    _REMOTE_ENDPOINT_SUB_TLV = {
        'as_number' : {
            'ixn_attr' : 'as4Number',
            'default' : '0'
        },
        'address_family' : {
            'ixn_attr' : 'addressFamily',
            'default': 'ipv4'
        },
        'ipv4_address' : {
            'ixn_attr' : 'remoteEndpointIPv4',
            'default': '0.0.0.0'
        },
        'ipv6_address': {
            'ixn_attr': 'remoteEndpointIPv6',
            'default': '::'
        }
    }
    
    _PREFERENCE_SUB_TLV = {
        'preference' : {
            'ixn_attr' : 'prefValue',
            'default': '0'
        }
    }
    
    _BINDING_SUB_TLV = {
        'binding_sid_type' : {
            'ixn_attr' : 'bindingSIDType',
            'default' : 'nobinding',
            'enum_map' : {
                'no_binding' : 'nobinding',
                'four_octet_sid' : 'sid4',
                'ipv6_sid' : 'ipv6sid'
            }
        },
        'four_octet_sid' : {
            'ixn_attr' : 'SID4Octet',
            'default': '0'
        },
        'bsid_as_mpls_label' : {
            'ixn_attr' : 'useAsMPLSLabel',
            'default': False
        },
        'ipv6_sid' : {
            'ixn_attr' : 'IPv6SID',
            'default': '::'
        },
        's_flag' : {
            'ixn_attr' : 'sflag',
            'default': False
        },
        'i_flag' : {
            'ixn_attr': 'iflag',
            'default': False
        },
        'remaining_flag_bits' : {
            'ixn_attr': 'remainingBits',
            'default': '0x01'
        }
    }

    _ENLP_SUB_TLV = {
        'explicit_null_label_policy' : {
            'ixn_attr' : 'ENLPValue',
            'default' : '4',
            'enum_map' : {
                'reserved_enlp' : '0',
                'push_ipv4_enlp' : '1',
                'push_ipv6_enlp' : '2',
                'push_ipv4_ipv6_enlp' : '3',
                'do_not_push_enlp' : '4'
            }
        }
    }

    _POLICIES_SEGMENT_LIST = {
        'segment_weight' : {
            'ixn_attr' : 'weight',
            'default' : '200'
        }
    }

    _SEGMENTS = {
        'segment_type' : {
            'ixn_attr' : 'segmentType',
            'default' : 'mplssid',
            'enum_map' : {
                'mpls_sid' : 'mplssid',
                'ipv6_sid' : 'ipv6sid'
            }
        },
        'mpls_label' : {
            'ixn_attr' : 'label',
            'default' : '16'
        },
        'mpls_tc' : {
            'ixn_attr' : 'trafficClass',
            'default' : '0'
        },
        'mpls_ttl' : {
            'ixn_attr' : 'timeToLive',
            'default' : '255'
        },
        'v_flag' : {
            'ixn_attr' : 'vflag',
            'default': False
        },
        'remaining_flag_bits' : {
            'ixn_attr' : 'remainingBits',
            'default': '0x01'
        },
        'ipv6_sid' : {
            'ixn_attr' : 'ipv6SID',
            'default' : '::'
        }
    }

    def __init__(self, ngpf):
        self._ngpf = ngpf
        self._api = ngpf._api
        self.update = ngpf.update
        self.configure_value = ngpf.configure_value
        self.get_xpath = ngpf.get_xpath
        self.select_node = ngpf.select_node
        self.select_child_node = ngpf.select_child_node

    def configure_bgpv4(self, ixn_parent, bgpv4, ixn_dg):
        ixn_bgpv4 = ixn_parent.BgpIpv4Peer
        self._api._remove(ixn_bgpv4, [bgpv4])
        name = self._api.special_char(bgpv4.name)
        args = {
            'Name': name,
        }
        ixn_bgpv4.find(Name='^%s$' % name)
        if len(ixn_bgpv4) == 0:
            ixn_bgpv4.add(**args)[-1]
        else:
            self.update(ixn_bgpv4, **args)
        self._api.ixn_objects[bgpv4.name] = ixn_bgpv4.href
        as_type = 'internal'
        if bgpv4.as_type is not None and bgpv4.as_type \
                    == 'ebgp':
            as_type = 'external'
        bgp_xpath = self.get_xpath(ixn_bgpv4.href)
        self.configure_value(bgp_xpath, 'type', as_type)
        as_bytes = bgpv4.as_number_width
        if as_bytes is None or as_bytes == 'two':
            self.configure_value(bgp_xpath, 'localAs2Bytes', bgpv4.as_number)
        elif as_bytes == 'four':
            self.configure_value(bgp_xpath, 'enable4ByteAs', True)
            self.configure_value(bgp_xpath, 'localAs4Bytes', bgpv4.as_number)
        else:
            raise Exception("Please configure supported [two, four] as_number_width")
        self.configure_value(bgp_xpath, 'dutIp', bgpv4.dut_address)
        self.configure_value(bgp_xpath, 'asSetMode',
                           bgpv4.as_number_set_mode,
                           enum_map=ConfigureBgp._BGP_AS_SET_MODE)
        # self._configure_pattern(ixn_dg.RouterData.RouterId, bgpv4.router_id)
        advanced = bgpv4.advanced
        self.configure_value(bgp_xpath, 'holdTimer', advanced.hold_time_interval)
        self.configure_value(bgp_xpath, 'keepaliveTimer', advanced.keep_alive_interval)
        self.configure_value(bgp_xpath, 'md5Key', advanced.md5_key)
        self.configure_value(bgp_xpath, 'updateInterval', advanced.update_interval)
        self.configure_value(bgp_xpath, 'ttl', advanced.time_to_live)
        self._configure_sr_te(ixn_bgpv4, bgp_xpath, bgpv4.sr_te_policies)
        self._bgp_route_builder(ixn_dg, ixn_bgpv4, bgpv4)
        return ixn_bgpv4
    
    def _bgp_route_builder(self, ixn_dg, ixn_bgp, bgp):
        bgpv4_routes = bgp.bgpv4_routes
        bgpv6_routes = bgp.bgpv6_routes
        if len(bgpv4_routes) > 0:
            for route_range in bgpv4_routes:
                self._configure_bgpv4_route(ixn_dg,
                                            ixn_bgp,
                                            route_range)
        if len(bgpv6_routes) > 0:
            for route_range in bgpv6_routes:
                self._configure_bgpv6_route(ixn_dg,
                                            ixn_bgp,
                                            route_range)

    def _configure_bgpv4_route(self, ixn_dg, ixn_bgp, route_range):
        ixn_ng = ixn_dg.NetworkGroup
        name = self._api.special_char(route_range.name)
        args = {
            'Name': name,
        }
        ixn_ng.find(Name='^%s$' % name)
        if len(ixn_ng) == 0:
            self.stop_topology()
            ixn_ng.add(**args)[-1]
            ixn_pool = ixn_ng.Ipv4PrefixPools.add()
        else:
            self.update(ixn_ng, **args)
            ixn_pool = ixn_ng.Ipv4PrefixPools.find()
        ixn_pool.Connector.find().ConnectedTo = ixn_bgp.href
        if route_range.name is not None:
            self._api.ixn_objects[route_range.name] = ixn_ng.href
            self._api._device_encap[route_range.name] = 'ipv4'
        pool_infos = self.select_node(ixn_pool.href,
                   children = ['bgpIPRouteProperty', 'bgpV6IPRouteProperty'])
        pool_xpath = pool_infos['xpath']
        addresses = route_range.addresses
        if len(addresses) > 0:
            ixn_ng.Multiplier = len(addresses)
            route_addresses = RouteAddresses()
            for address in addresses:
                route_addresses.address = address.address
                route_addresses.step = address.step
                route_addresses.prefix = address.prefix
                route_addresses.count = address.count
            self.configure_value(pool_xpath, 'networkAddress',
                                  route_addresses.address)
            self.configure_value(pool_xpath, 'prefixAddrStep',
                                  route_addresses.step)
            self.configure_value(pool_xpath, 'prefixLength',
                                  route_addresses.prefix)
            self.configure_value(pool_xpath, 'numberOfAddressesAsy',
                                  route_addresses.count)
        if 'bgpIPRouteProperty' in pool_infos:
            ixn_bgp_property = ixn_pool.BgpIPRouteProperty.find()
            property_xpath = pool_infos['bgpIPRouteProperty'][0]['xpath']
        else:
            ixn_bgp_property = ixn_pool.BgpV6IPRouteProperty.find()
            property_xpath = pool_infos['bgpV6IPRouteProperty'][0]['xpath']
        self.configure_value(property_xpath, 'ipv4NextHop',
                              route_range.next_hop_address)
        if route_range.name is not None:
            ixn_bgp_property.Name = route_range.name
            self._api.ixn_route_objects[route_range.name] = ixn_bgp_property
        advanced = route_range.advanced
        if advanced.multi_exit_discriminator is not None:
            self.configure_value(property_xpath, 'enableMultiExitDiscriminator', True)
            self.configure_value(property_xpath, 'multiExitDiscriminator',
                                  advanced.multi_exit_discriminator)
        self.configure_value(property_xpath, 'origin', advanced.origin)
        self._config_bgp_as_path(route_range.as_path, ixn_bgp_property)
        self._config_bgp_community(route_range.communities, ixn_bgp_property)

    def configure_bgpv6(self, ixn_parent, bgpv6, ixn_dg):
        ixn_bgpv6 = ixn_parent.BgpIpv6Peer
        self._api._remove(ixn_bgpv6, [bgpv6])
        name = self._api.special_char(bgpv6.name)
        args = {
            'Name': name,
        }
        ixn_bgpv6.find(Name='^%s$' % name)
        if len(ixn_bgpv6) == 0:
            ixn_bgpv6.add(**args)[-1]
        else:
            self.update(ixn_bgpv6, **args)
        self._api.ixn_objects[bgpv6.name] = ixn_bgpv6.href
        as_type = 'internal'
        if bgpv6.as_type is not None and bgpv6.as_type \
            == 'ebgp':
            as_type = 'external'
        bgp_xpath = self.get_xpath(ixn_bgpv6.href)
        self.configure_value(bgp_xpath, 'type', as_type)
        as_bytes = bgpv6.as_number_width
        if as_bytes is None or as_bytes == 'two':
            self.configure_value(bgp_xpath, 'localAs2Bytes', bgpv6.as_number)
        elif as_bytes == 'four':
            self.configure_value(bgp_xpath, 'enable4ByteAs', True)
            self.configure_value(bgp_xpath, 'localAs4Bytes', bgpv6.as_number)
        else:
            raise Exception("Please configure supported [two, four] as_number_width")
        self.configure_value(bgp_xpath, 'dutIp', bgpv6.dut_address)
        self.configure_value(bgp_xpath, 'asSetMode',
                             bgpv6.as_number_set_mode,
                             enum_map=ConfigureBgp._BGP_AS_SET_MODE)
        # self._configure_pattern(ixn_dg.RouterData.RouterId, bgpv4.router_id)
        advanced = bgpv6.advanced
        self.configure_value(bgp_xpath, 'holdTimer', advanced.hold_time_interval)
        self.configure_value(bgp_xpath, 'keepaliveTimer', advanced.keep_alive_interval)
        self.configure_value(bgp_xpath, 'md5Key', advanced.md5_key)
        self.configure_value(bgp_xpath, 'updateInterval', advanced.update_interval)
        self.configure_value(bgp_xpath, 'ttl', advanced.time_to_live)
        self._configure_sr_te(ixn_bgpv6, bgp_xpath, bgpv6.sr_te_policies)
        self._bgp_route_builder(ixn_dg, ixn_bgpv6, bgpv6)
        return ixn_bgpv6

    def _configure_bgpv6_route(self, ixn_dg, ixn_bgp, route_range):
        ixn_ng = ixn_dg.NetworkGroup
        name = self._api.special_char(route_range.name)
        args = {
            'Name': name,
        }
        ixn_ng.find(Name='^%s$' % name)
        if len(ixn_ng) == 0:
            self.stop_topology()
            ixn_ng.add(**args)[-1]
            ixn_pool = ixn_ng.Ipv6PrefixPools.add()
        else:
            self.update(ixn_ng, **args)
            ixn_pool = ixn_ng.Ipv6PrefixPools.find()
        ixn_pool.Connector.find().ConnectedTo = ixn_bgp.href
        if route_range.name is not None:
            self._api.ixn_objects[route_range.name] = ixn_ng.href
            self._api._device_encap[route_range.name] = 'ipv6'
        pool_infos = self.select_node(ixn_pool.href,
                                      children=['bgpIPRouteProperty', 'bgpV6IPRouteProperty'])
        pool_xpath = pool_infos['xpath']
        addresses = route_range.addresses
        if len(addresses) > 0:
            ixn_ng.Multiplier = len(addresses)
            route_addresses = RouteAddresses()
            for address in addresses:
                route_addresses.address = address.address
                route_addresses.step = address.step
                route_addresses.prefix = address.prefix
                route_addresses.count = address.count
            self.configure_value(pool_xpath, 'networkAddress',
                                 route_addresses.address)
            self.configure_value(pool_xpath, 'prefixAddrStep',
                                 route_addresses.step)
            self.configure_value(pool_xpath, 'prefixLength',
                                 route_addresses.prefix)
            self.configure_value(pool_xpath, 'numberOfAddressesAsy',
                                 route_addresses.count)
        if self._api.get_device_encap(ixn_dg.Name) == 'ipv4':
            ixn_bgp_property = ixn_pool.BgpIPRouteProperty.find()
            property_xpath = pool_infos['bgpIPRouteProperty'][0]['xpath']
        else:
            ixn_bgp_property = ixn_pool.BgpV6IPRouteProperty.find()
            property_xpath = pool_infos['bgpV6IPRouteProperty'][0]['xpath']
        self.configure_value(property_xpath, 'ipv6NextHop',
                             route_range.next_hop_address)
        if route_range.name is not None:
            ixn_bgp_property.Name = route_range.name
            self._api.ixn_route_objects[route_range.name] = ixn_bgp_property
        advanced = route_range.advanced
        if advanced.multi_exit_discriminator is not None:
            self.configure_value(property_xpath, 'enableMultiExitDiscriminator', True)
            self.configure_value(property_xpath, 'multiExitDiscriminator',
                                 advanced.multi_exit_discriminator)
        self.configure_value(property_xpath, 'origin', advanced.origin)
        self._config_bgp_as_path(route_range.as_path, ixn_bgp_property)
        self._config_bgp_community(route_range.communities, ixn_bgp_property)

    def _config_bgp_as_path(self, as_path, ixn_bgp_property):
        as_path_segments = as_path.as_path_segments
        property_xpath = self.get_xpath(ixn_bgp_property.href)
        if as_path.as_set_mode is not None or len(
            as_path_segments) > 0:
            self.configure_value(property_xpath, 'enableAsPathSegments', True)
            self.configure_value(property_xpath, 'asSetMode', as_path.as_set_mode,
                                 enum_map=ConfigureBgp._BGP_AS_MODE)
            self.configure_value(property_xpath, 'OverridePeerAsSetMode',
                                 as_path.override_peer_as_set_mode)
            if len(as_path_segments) > 0:
                ixn_bgp_property.NoOfASPathSegmentsPerRouteRange = len(
                    as_path_segments)
                ixn_segments = ixn_bgp_property.BgpAsPathSegmentList.find()
                for seg_index, segment in enumerate(as_path_segments):
                    ixn_segment = ixn_segments[seg_index]
                    ixn_segment.SegmentType.Single(ConfigureBgp._BGP_SEG_TYPE[
                                                       segment.segment_type])
                    as_numbers = segment.as_numbers
                    if as_numbers is not None:
                        ixn_segment.NumberOfAsNumberInSegment = len(as_numbers)
                        as_numbers_info = self.select_child_node(ixn_segment.href,
                                                                 'bgpAsNumberList')
                        for as_index, as_number in enumerate(as_numbers):
                            as_num_xpath = as_numbers_info[as_index]['xpath']
                            self.configure_value(as_num_xpath, 'asNumber', as_number)

    def _config_bgp_community(self, communities, ixn_bgp_property):
        if len(communities) == 0:
            ixn_bgp_property.EnableCommunity.Single(False)
            return
        ixn_bgp_property.EnableCommunity.Single(True)
        ixn_bgp_property.NoOfCommunities = len(communities)
        communities_info = self.select_child_node(ixn_bgp_property.href,
                                                  'bgpCommunitiesList')
        for index, community in enumerate(communities):
            community_xpath = communities_info[index]['xpath']
            if community.community_type is not None:
                self.configure_value(community_xpath, 'type', community.community_type,
                                     enum_map=ConfigureBgp._BGP_COMMUNITY_TYPE)
            self.configure_value(community_xpath, 'asNumber', community.as_number)
            self.configure_value(community_xpath, 'lastTwoOctets', community.as_custom)

    def _configure_sr_te(self, ixn_bgp, bgp_xpath, sr_te_list):
        if sr_te_list is None or len(sr_te_list) == 0:
            return
        self.configure_value(bgp_xpath, 'capabilitySRTEPoliciesV4', True)
        self.configure_value(bgp_xpath, 'capabilitySRTEPoliciesV6', True)
        ixn_bgp.NumberSRTEPolicies = len(sr_te_list)
        if re.search('bgpIpv4Peer', ixn_bgp.href) is not None:
            ixn_sr_te = ixn_bgp.BgpSRTEPoliciesListV4
        else:
            ixn_sr_te = ixn_bgp.BgpSRTEPoliciesListV6
        sr_te_xpath = self.get_xpath(ixn_sr_te.href)
        self._configure_attributes(ConfigureBgp._BGP_SR_TE,
                                   sr_te_list,
                                   sr_te_xpath)
        next_hops = []
        add_paths = []
        as_paths = []
        communities = []
        for sr_te in sr_te_list:
            next_hops.append(sr_te.next_hop)
            add_paths.append(sr_te.add_path)
            as_paths.append(sr_te.as_path)
            communities.append(sr_te.communities)
        active_list = self._process_nodes(next_hops)
        self.configure_value(sr_te_xpath, 'enableNextHop', active_list)
        if any(active_list):
            self._configure_attributes(ConfigureBgp._SRTE_NEXT_HOP,
                                       next_hops,
                                       sr_te_xpath)
        active_list = self._process_nodes(add_paths)
        self.configure_value(sr_te_xpath, 'enableAddPath', active_list)
        if any(active_list):
            self._configure_attributes(ConfigureBgp._SRTE_ADDPATH,
                                       add_paths,
                                       sr_te_xpath)
        active_list = self._process_nodes(as_paths)
        if any(active_list):
            self._configure_attributes(ConfigureBgp._SRTE_AS_PATH,
                                       as_paths,
                                       sr_te_xpath)
            self._configure_srte_aspath_segment(as_paths,
                                                ixn_sr_te)
        self._configure_tlvs(ixn_sr_te, sr_te_list)
    
    def _get_symmetric_nodes(self, parent_list, node_name):
        NodesInfo = namedtuple("NodesInfo", ["max_len",
                                             "active_list",
                                             "symmetric_nodes"])
        nodes_list = []
        max_len = 0
        for parent in parent_list:
            nodes = getattr(parent, node_name)
            node_len = len(nodes)
            if node_len > max_len:
                max_len = node_len
            nodes_list.append(nodes)
        symmetric_nodes = []
        active_list = []
        for nodes in nodes_list:
            if len(nodes) == max_len:
                for node in nodes:
                    active_list.append(node.active)
                    symmetric_nodes.append(node)
            else:
                for index in range(0, max_len):
                    node = nodes[0]
                    if index < len(nodes):
                        node = nodes[index]
                        active_list.append(node.active)
                        symmetric_nodes.append(node)
                    else:
                        active_list.append(False)
                        symmetric_nodes.append(node)
        return NodesInfo(max_len,
                         active_list,
                         symmetric_nodes)
    
    def _get_symetric_tab_nodes(self, parent_list, node_name):
        TabNodesInfo = namedtuple("TabNodesInfo", ["max_len",
                                           "symmetric_nodes_list",
                                           "actives_list"])
        max_len = 0
        symmetric_nodes_list = []
        actives_list = []
        is_enable = False
        for parent in parent_list:
            nodes = getattr(parent, node_name)
            if nodes is None:
                continue
            is_enable = True
            node_len = len(nodes)
            if node_len > max_len:
                for index in range(max_len, node_len):
                    symmetric_nodes_list.append([nodes[index]] * len(
                                parent_list))
                    actives_list.append([False] * len(parent_list))
                max_len = node_len
        if is_enable:
            for parent_idx, parent in enumerate(parent_list):
                nodes = getattr(parent, node_name)
                for node_idx, node in enumerate(nodes):
                    symmetric_nodes_list[node_idx][parent_idx] = node
                    actives_list[node_idx][parent_idx] = True
        return TabNodesInfo(max_len,
                            symmetric_nodes_list,
                            actives_list)

    def _configure_srte_aspath_segment(self, as_paths, ixn_sr_te):
        nodes_list_info = self._get_symetric_tab_nodes(as_paths, 'as_path_segments')
        if nodes_list_info.max_len == 0:
            return
        ixn_sr_te.EnableAsPathSegments.Single(True)
        ixn_sr_te.NoOfASPathSegmentsPerRouteRange = nodes_list_info.max_len
        ixn_segments = ixn_sr_te.BgpAsPathSegmentList
        ixn_segments.find()
        segments_info = self.select_node(ixn_sr_te.refresh().href,
                                         children=['bgpAsPathSegmentList'])
        for seg_idx , segment in enumerate(segments_info['bgpAsPathSegmentList']):
            segment_xpath = segment['xpath']
            ixn_segment = ixn_segments[seg_idx]
            segment_nodes = nodes_list_info.symmetric_nodes_list[seg_idx]
            self.configure_value(segment_xpath, 'enableASPathSegment',
                                 nodes_list_info.actives_list[seg_idx])
            self._configure_attributes(ConfigureBgp._SRTE_ASPATH_SEGMENT,
                                       segment_nodes,
                                       segment_xpath)
            configure_as_number = False
            for segment_node in segment_nodes:
                as_numbers = getattr(segment_node, 'as_numbers')
                if as_numbers is not None:
                    configure_as_number = True
                    if not isinstance(as_numbers, list):
                        raise Exception("as_numbers must be list")
            if configure_as_number is True:
                as_numbers_list = self._get_symetric_tab_nodes(segment_nodes, 'as_numbers')
                ixn_segment.NumberOfAsNumberInSegment = as_numbers_list.max_len
                numbers_info = self.select_node(ixn_segment.href,
                                         children=['bgpAsNumberList'])
                for num_idx, number in enumerate(numbers_info['bgpAsNumberList']):
                    number_xpath = number['xpath']
                    self.configure_value(number_xpath, 'enableASNumber',
                                         as_numbers_list.actives_list[num_idx])
                    self.configure_value(number_xpath, 'asNumber',
                                         as_numbers_list.symmetric_nodes_list[num_idx])
        
    def _configure_tlvs(self, ixn_sr_te, sr_te_list):
        nodes_info = self._get_symmetric_nodes(sr_te_list, 'tunnel_tlvs')
        if int(nodes_info.max_len) > 2:
            raise Exception("Value {0} for SR TE Policy Number of Tunnel TLVs is "
                            "greater than maximal value 2".format(nodes_info.max_len))
        if re.search('bgpSRTEPoliciesListV4', ixn_sr_te.href) is not None:
            ixn_sr_te.NumberOfTunnelsV4 = nodes_info.max_len
            ixn_tunnel = ixn_sr_te.BgpSRTEPoliciesTunnelEncapsulationListV4
        else:
            ixn_sr_te.NumberOfTunnelsV6 = nodes_info.max_len
            ixn_tunnel = ixn_sr_te.BgpSRTEPoliciesTunnelEncapsulationListV6
        tunnel_xpath = self.get_xpath(ixn_tunnel.href)
        self.configure_value(tunnel_xpath, 'active', nodes_info.active_list)
        tunnel_tlvs = nodes_info.symmetric_nodes
        
        remote_endpoint_sub_tlv = []
        preference_sub_tlv = []
        binding_sub_tlv = []
        explicit_null_label_policy_sub_tlv = []
        for tunnel_tlv in tunnel_tlvs:
            remote_endpoint_sub_tlv.append(tunnel_tlv.remote_endpoint_sub_tlv)
            preference_sub_tlv.append(tunnel_tlv.preference_sub_tlv)
            binding_sub_tlv.append(tunnel_tlv.binding_sub_tlv)
            explicit_null_label_policy_sub_tlv.append(tunnel_tlv.explicit_null_label_policy_sub_tlv)
        active_list = self._process_nodes(remote_endpoint_sub_tlv)
        self.configure_value(tunnel_xpath, 'enRemoteEndPointTLV', active_list)
        if any(active_list):
            self._configure_attributes(ConfigureBgp._REMOTE_ENDPOINT_SUB_TLV,
                                       remote_endpoint_sub_tlv,
                                       tunnel_xpath)
        active_list = self._process_nodes(preference_sub_tlv)
        self.configure_value(tunnel_xpath, 'enPrefTLV', active_list)
        if any(active_list):
            self._configure_attributes(ConfigureBgp._PREFERENCE_SUB_TLV,
                                       preference_sub_tlv,
                                       tunnel_xpath)
        active_list = self._process_nodes(binding_sub_tlv)
        self.configure_value(tunnel_xpath, 'enBindingTLV', active_list)
        if any(active_list):
            self._configure_attributes(ConfigureBgp._BINDING_SUB_TLV,
                                       binding_sub_tlv,
                                       tunnel_xpath)
        active_list = self._process_nodes(explicit_null_label_policy_sub_tlv)
        self.configure_value(tunnel_xpath, 'enENLPTLV', active_list)
        if any(active_list):
            self._configure_attributes(ConfigureBgp._ENLP_SUB_TLV,
                                       explicit_null_label_policy_sub_tlv,
                                       tunnel_xpath)
        self._configure_tlv_segment(ixn_tunnel, tunnel_tlvs)

    def _configure_tlv_segment(self, ixn_tunnel, tunnel_tlvs):
        nodes_info = self._get_symmetric_nodes(tunnel_tlvs, 'segment_lists')
        if re.search('bgpSRTEPoliciesTunnelEncapsulationListV4', ixn_tunnel.href) is not None:
            ixn_tunnel.NumberOfSegmentListV4 = nodes_info.max_len
            ixn_segment_list = ixn_tunnel.BgpSRTEPoliciesSegmentListV4
        else:
            ixn_tunnel.NumberOfSegmentListV6 = nodes_info.max_len
            ixn_segment_list = ixn_tunnel.BgpSRTEPoliciesSegmentListV6
        segment_list_xpath = self.get_xpath(ixn_segment_list.href)
        self.configure_value(segment_list_xpath, 'active', nodes_info.active_list)
        segment_list = nodes_info.symmetric_nodes
        if any(nodes_info.active_list):
            self._configure_attributes(ConfigureBgp._POLICIES_SEGMENT_LIST,
                                       segment_list,
                                       segment_list_xpath)
            self.configure_value(segment_list_xpath, 'enWeight',
                                 [True] * len(segment_list))
        nodes_info = self._get_symmetric_nodes(segment_list, 'segments')
        if re.search('bgpSRTEPoliciesSegmentListV4', ixn_segment_list.href) is not None:
            ixn_segment_list.NumberOfSegmentsV4 = nodes_info.max_len
            ixn_segments = ixn_segment_list.BgpSRTEPoliciesSegmentsCollectionV4
        else:
            ixn_segment_list.NumberOfSegmentsV6 = nodes_info.max_len
            ixn_segments = ixn_segment_list.BgpSRTEPoliciesSegmentsCollectionV6
        segments_xpath = self.get_xpath(ixn_segments.href)
        self.configure_value(segments_xpath, 'active', nodes_info.active_list)
        segments = nodes_info.symmetric_nodes
        if any(nodes_info.active_list):
            self._configure_attributes(ConfigureBgp._SEGMENTS,
                                       segments,
                                       segments_xpath)
    
    def _process_nodes(self, nodes):
        active_list = []
        for index, node in enumerate(nodes):
            active = False
            if node is None:
                if index == 0:
                    nodes[0] = next(v for v in nodes if v is not None)
                else:
                    nodes[index] = nodes[index - 1]
            else:
                is_config = False
                for name, value in node._properties.items():
                    if value is not None:
                        is_config = True
                        break
                if is_config is True:
                    active = True
            active_list.append(active)
        return active_list
    
    def _configure_attributes(self, mapper, parent_list, xpath):
        for attribute in mapper:
            attr_mapper = mapper[attribute]
            ixn_attribute = attr_mapper['ixn_attr']
            default_value = attr_mapper['default']
            enum_map = attr_mapper.get('enum_map')
            default_obj = getattr(self, str(default_value), None)
            config_values = []
            for parent in parent_list:
                config_value = getattr(parent, attribute, None)
                if config_value is not None:
                    if enum_map is None:
                        config_values.append(str(config_value))
                    else:
                        if str(config_value) not in enum_map.keys():
                            raise Exception("{0} must configure with enum {1}".format(
                                        attribute, enum_map.keys()))
                        config_values.append(enum_map[
                            str(config_value)])
                elif default_obj is not None:
                    config_values.append(default_obj())
                else:
                    config_values.append(default_value)
            self.configure_value(xpath, ixn_attribute, config_values)
    
    def stop_topology(self):
        glob_topo = self._api._globals.Topology.refresh()
        if glob_topo.Status == 'started':
            self._api._ixnetwork.StopAllProtocols('sync')

class RouteAddresses(object):
    def __init__(self):
        self._address = []
        self._count = []
        self._prefix = []
        self._step = []
    
    @property
    def address(self):
        return self._address
    
    @address.setter
    def address(self, value):
        self._address.append(value)
    
    @property
    def count(self):
        return self._count
    
    @count.setter
    def count(self, value):
        self._count.append(value)
    
    @property
    def prefix(self):
        return self._prefix
    
    @prefix.setter
    def prefix(self, value):
        self._prefix.append(value)
    
    @property
    def step(self):
        return self._step
    
    @step.setter
    def step(self, value):
        self._step.append(value)


from copy import deepcopy

class CustomField(object):
    """
    Implemented All custom field which specify within TrafficItem class
    Best Practice :
        - Please use self._configure_pattern to setting Pattern
    """
    _IPv4_DSCP_PHB = {
        '0': 'ipv4.header.priority.ds.phb.defaultPHB.defaultPHB',
        '8': 'ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB',
        '16': 'ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB',
        '24': 'ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB',
        '32': 'ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB',
        '40': 'ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB',
        '48': 'ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB',
        '56': 'ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB',
        '10': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '12': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '14': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '18': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '20': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '22': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '26': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '28': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '30': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '34': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '36': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '38': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB',
        '46': 'ipv4.header.priority.ds.phb.expeditedForwardingPHB.expeditedForwardingPHB'
    }
    
    _IPv4_DSCP_ECN = {
        'ipv4.header.priority.ds.phb.defaultPHB.defaultPHB': 'ipv4.header.priority.ds.phb.defaultPHB.unused',
        'ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB': 'ipv4.header.priority.ds.phb.classSelectorPHB.unused',
        'ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB': 'ipv4.header.priority.ds.phb.assuredForwardingPHB.unused',
        'ipv4.header.priority.ds.phb.expeditedForwardingPHB.expeditedForwardingPHB': 'ipv4.header.priority.ds.phb.expeditedForwardingPHB.unused'
    }
    
    _ETHERNETPAUSE = {
        'dst': 'ethernet.header.destinationAddress',
        'src': 'ethernet.header.sourceAddress',
        'ether_type': 'ethernet.header.etherType',
        'control_op_code': 'custom.header.length',
        'time': 'custom.header.data',
    }

    def _ipv4_priority(self, ixn_field, priority):
        if priority.choice == 'dscp':
            field_type_id = None
            first_value = str(self._get_first_value(priority.dscp.phb))
            if first_value in CustomField._IPv4_DSCP_PHB:
                field_type_id = CustomField._IPv4_DSCP_PHB[first_value]
            else:
                field_type_id = CustomField._IPv4_DSCP_PHB["0"]
            if field_type_id is not None:
                self._configure_pattern(ixn_field, field_type_id, priority.dscp.phb, field_choice=True)
                field_type_id = CustomField._IPv4_DSCP_ECN[field_type_id]
                self._configure_pattern(ixn_field, field_type_id, priority.dscp.ecn, field_choice=True)
        elif priority.choice == 'tos':
            self._configure_field(ixn_field, priority, field_choice=True)
        else:
            self._configure_pattern(ixn_field, 'ipv4.header.priority.raw', priority.raw, field_choice=True)

    def _get_first_value(self, phb_pattern):
        if phb_pattern.choice is None:
            return None
        if phb_pattern.choice == 'value':
            value = phb_pattern.value
        elif phb_pattern.choice == 'values':
            value = phb_pattern.values[0]
        elif phb_pattern.choice == 'counter':
            value = phb_pattern.counter.start
        else:
            value = phb_pattern.random.min
        return value
    
    def adjust_header(self, headers):
        ''''''
        new_headers = list()
        for header in headers:
            header = header.parent
            if header.choice == 'ethernetpause':
                self._ethernet_pause(
                    new_headers, headers, header.ethernetpause
                )
            elif header.choice == 'gtpv1':
                self._gtpv1(new_headers, header)
            else:
                new_headers.append(header)
        return new_headers

    def _ethernet_pause(self, new_headers, headers, ethernetpause):
        ''' Current IxNetwork 9.10 do not have Global Pause support.
        Therefore using Ether > Custom header as work around.
        We will remove these code and add EthernetPause in generic section.
        The implementation will only support fixed patterns for
        control_op_code and time
        '''
        ethpause = deepcopy(ethernetpause)
        headers.clear()
        eth = headers.ethernet()[-1]
        cust = headers.custom()[-1]
        new_headers.append(eth.parent)
        new_headers.append(cust.parent)
        if ethpause._properties.get('src') is not None:
            eth.src.deserialize(ethpause.src.serialize('dict'))
        else:
            eth.src.value = '00:00:aa:00:00:01'
        if ethpause._properties.get('dst') is not None:
            eth.dst.deserialize(ethpause.dst.serialize('dict'))
        else:
            eth.dst.value = '01:80:c2:00:00:01'
        if ethpause._properties.get('ether_type') is not None:
            eth.ether_type.deserialize(ethpause.ether_type.serialize('dict'))
        else:
            eth.ether_type.value = '8808'
        if ethpause._properties.get('control_op_code') is not None:
            control_op_code =\
                self._get_first_value(ethpause.control_op_code).zfill(4)
        else:
            control_op_code = '0001'
        if ethpause._properties.get('time') is not None:
            time = self._get_first_value(ethpause.time).zfill(4)
        else:
            time = 'FFFF'
        cust.bytes = '{0}{1}'.format(control_op_code, time)

    def _gtpv1(self, new_headers, header):
        ''''''
        new_headers.append(header)
        gtp_option = deepcopy(header)
        gtp_option.__setattr__('choice', 'gtpv1option')
        gtp_option.__setattr__('gtpv1option', header.gtpv1)
        gtp_option.__delattr__('gtpv1')
        new_headers.append(gtp_option)
    
    def _custom_headers(self, ixn_field, packet):
        if packet.bytes is not None:
            ixn_custom_length = ixn_field.find(FieldTypeId='custom.header.length')
            ixn_custom_length.update(Auto=False, ValueType='singleValue', SingleValue=len(packet.bytes) * 4)
            ixn_custom_data = ixn_field.find(FieldTypeId='custom.header.data')
            ixn_custom_data.update(Auto=False, ValueType='singleValue', SingleValue=packet.bytes)


import json
import time, re
from collections import namedtuple
from ixnetwork_restpy import TestPlatform, SessionAssistant
import snappi
from snappi_ixnetwork.validation import Validation
from snappi_ixnetwork.vport import Vport
from snappi_ixnetwork.lag import Lag
from snappi_ixnetwork.ngpf import Ngpf
from snappi_ixnetwork.trafficitem import TrafficItem
from snappi_ixnetwork.capture import Capture
from snappi_ixnetwork.timer import Timer
from snappi_ixnetwork.protocolmetrics import ProtocolMetrics
from snappi_ixnetwork.resourcegroup import ResourceGroup
from snappi_ixnetwork.exceptions import SnappiIxnException

class Api(snappi.Api):
    """IxNetwork implementation of the abstract-open-traffic-generator package

    Args
    ----
    - address (str): The address of the IxNetwork API Server
    - port (str): The rest port of the IxNetwork API Server
    - username (str): The username for Linux IxNetwork API Server
        This is not required when connecting to single session environments
    - password (str): The password for Linux IxNetwork API Server
        This is not required when connecting to single session environments
    """
    def __init__(self,
                 **kwargs):
        """Create a session
        - address (str): The ip address of the TestPlatform to connect to
        where test sessions will be created or connected to.
        - port (str): The rest port of the TestPlatform to connect to.
        - username (str): The username to be used for authentication
        - password (str): The password to be used for authentication
        """
        super(Api, self).__init__()
        host = kwargs.get('host')
        location = kwargs.get('location')
        username = kwargs.get('username')
        password = kwargs.get('password')
        license_servers = kwargs.get('license_servers')
        log_level = kwargs.get('loglevel')
        # temporary fix to allow tests to run
        if host is not None:
            location = host
        location = 'https://127.0.0.1:11009' if location is None else location
        self._address, self._port = self._get_addr_port(location)
        self._username = 'admin' if username is None else username
        self._password = 'admin' if password is None else password
        self._license_servers = [] if license_servers is None else license_servers
        self._log_level = 'info'
        self._running_config = None
        self._config = None
        self._assistant = None
        self._ixn_errors = list()
        self._config_objects = {}
        self._device_encap = {}
        self._config_type = self.config()
        self._transmit_state = self.transmit_state()
        self._link_state = self.link_state()
        self._capture_state = self.capture_state()
        self._capture_request = self.capture_request()
        self._ixn_route_objects = {}
        self.validation = Validation(self)
        self.vport = Vport(self)
        self.lag = Lag(self)
        self.ngpf = Ngpf(self)
        self.traffic_item = TrafficItem(self)
        self.capture = Capture(self)
        self.protocol_metrics = ProtocolMetrics(self)
        self.resource_group = ResourceGroup(self)

    def _get_addr_port(self, host):
        items = host.split('/')
        items = items[-1].split(':')

        addr = items[0]
        if len(items) == 2:
            return addr, items[-1]
        else:
            if host.startswith('https'):
                return addr, '443'
            else:
                return addr, '80'

    @property
    def snappi_config(self):
        return self._config

    def get_config_object(self, name):
        try:
            return self._config_objects[name]
        except KeyError:
            raise NameError("snappi object named {0} not found"
                            .format(name))

    def get_device_encap(self, name):
        try:
            return self._device_encap[name]
        except KeyError:
            raise NameError("snappi object named {0} not found"
                            .format(name))

    @property
    def ixn_objects(self):
        """A dict of all model unique names to ixn hrefs
        """
        return self._ixn_objects

    def get_ixn_object(self, name):
        """Returns an ixnetwork_restpy object given a unique configuration name
        """
        href = self.get_ixn_href(name)
        return self._assistant.Session.GetObjectFromHref(href)

    def get_ixn_href(self, name):
        """Returns an href given a unique configuration name
        """
        return self._ixn_objects[name]
    
    @property
    def ixn_route_objects(self):
        return self._ixn_route_objects
    
    def get_route_object(self, name):
        if name not in self._ixn_route_objects.keys():
            raise Exception("%s not within configure routes" %name)
        return self._ixn_route_objects[name]

    @property
    def assistant(self):
        return self._assistant

    def _dict_to_obj(self, source):
        """Returns an object given a dict
        """
        if isinstance(source, list):
            source = [self._dict_to_obj(x) for x in source]
        if not isinstance(source, dict):
            return source
        o = lambda: None
        for k, v in source.items():
            o.__dict__[k] = self._dict_to_obj(v)
        return o

    def _request_detail(self):
        request_detail = snappi.ResponseWarning()
        errors = self._errors
        warnings = list()
        app_errors = self._globals.AppErrors.find()
        if len(app_errors) > 0:
            current_errors = app_errors[0].Error.find()
            if len(current_errors) > 0:
                for error in current_errors:
                    # Add loop as sequence are not sorted
                    match = [o for o in self._ixn_errors if o.Name == error.Name
                                                    and o.LastModified == error.LastModified]
                    if len(match) == 0:
                        if error.ErrorLevel == 'kWarning':
                            warnings.append("IxNet - {0}".format(error.Description))
                        if error.ErrorLevel == 'kError':
                            errors.append("IxNet - {0}".format(error.Description))
        # request_detail.errors = errors
        if len(errors) > 0:
            self._errors = []
            raise SnappiIxnException(500, errors)
        request_detail.warnings = warnings
        return request_detail

    def set_config(self, config):
        """Set or update the configuration
        """
        try:
            if isinstance(config, (type(self._config_type),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[Config, str]')

            if isinstance(config, str) is True:
                config = self._config_type.deserialize(config)
            self.config_ixnetwork(config)
        except Exception as err:
            raise SnappiIxnException(err)
        app_errors = self._globals.AppErrors.find()
        bad_requests = []
        if len(app_errors) > 0:
            current_errors = app_errors[0].Error.find()
            if len(current_errors) > 0:
                for error in current_errors:
                    if error.Name == 'JSON Import Issues':
                        bad_requests = [instance.SourceValues[0]
                                        for instance in error.Instance.find()]
        if bad_requests:
            if len(bad_requests) == 1:
                raise SnappiIxnException(400, "Bad request error: {}".format(
                    bad_requests[0]))
            raise SnappiIxnException(400, "Bad request errors:", bad_requests)
        return self._request_detail()

    def config_ixnetwork(self, config):
        self._config_objects = {}
        self._device_encap = {}
        self._ixn_objects = {}
        self._connect()
        self.capture.reset_capture_request()
        self._config = self._validate_instance(
            config)
        with Timer(self, 'Config validation'):
            self.validation.validate_config()
        self._ixnetwork.Traffic.UseRfc5952 = True
        if len(self._config._properties) == 0:
            self._ixnetwork.NewConfig()
        else:
            self.vport.config()
            self.lag.config()
            with Timer(self, 'Devices configuration'):
                self.ngpf.config()
            with Timer(self, 'Flows configuration'):
                self.traffic_item.config()
        self._running_config = self._config
        self._apply_change()
    
    def set_transmit_state(self, payload):
        """Set the transmit state of flows
        """
        try:
            if isinstance(payload, (type(self._transmit_state),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[TransmitState, str]')
            if isinstance(payload, str) is True:
                payload = self._transmit_state.deserialize(
                                        payload)
            self._connect()
            self.traffic_item.transmit(payload)
        except Exception as err:
            raise SnappiIxnException(err)
        return self._request_detail()

    def set_link_state(self, link_state):
        try:
            if isinstance(link_state, (type(self._link_state),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[LinkState, str]')
            if isinstance(link_state, str):
                link_state = self._link_state.deserialize(link_state)
            self._connect()
            if link_state.port_names is not None:
                self.vport.set_link_state(link_state)
        except Exception as err:
            raise SnappiIxnException(err)
        return self._request_detail()
    
    def set_capture_state(self, payload):
        """Starts capture on all ports that have capture enabled.
        """
        try:
            if isinstance(payload, (type(self._capture_state),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[CaptureState, str]')
            if isinstance(payload, str) is True:
                payload = self._capture_state.deserialize(
                                        payload)
            self._connect()
            self.capture.set_capture_state(payload)
        except Exception as err:
            raise SnappiIxnException(err)
        return self._request_detail()

    def set_route_state(self, payload):
        try:
            route_state = self.route_state()
            if isinstance(payload, (type(route_state),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[RouteState, str]')
            if isinstance(payload, str) is True:
                payload = route_state.deserialize(
                                        payload)
            self._connect()
            with Timer(self, 'Setting route state'):
                self.ngpf.set_route_state(payload)
            return self._request_detail()
        except Exception as err:
            raise SnappiIxnException(err)

    def get_capture(self, request):
        """Gets capture file and returns it as a byte stream
        """
        try:
            if isinstance(request, (type(self._capture_request),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[CaptureRequest, str]')
            if isinstance(request, str) is True:
                request = self._capture_request.deserialize(
                    request)
            self._connect()
        except Exception as err:
            raise SnappiIxnException(err)
        return self.capture.results(request)

    def get_metrics(self, request):
        """
        Gets port, flow and protocol metrics.

        Args
        ----
        - request (Union[MetricsRequest, str]): A request for Port, Flow and
          protocol metrics.
          The request content MUST be vase on the OpenAPI model,
          #/components/schemas/Result.MetricsRequest
          See the docs/openapi.yaml document for all model details
        """
        try:
            self._connect()
            metric_req = self.metrics_request()
            if isinstance(request, (type(metric_req),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[MetricsRequest, str]')
            if isinstance(request, str) is True:
                request = metric_req.deserialize(request)
            # Need to change the code style when the choice Enum grows big
            if request.choice == 'port':
                response = self.vport.results(request.port)
                metric_res = self.metrics_response()
                metric_res.port_metrics.deserialize(response)
                return metric_res
            if request.choice == 'flow':
                response = self.traffic_item.results(request.flow)
                metric_res = self.metrics_response()
                metric_res.flow_metrics.deserialize(response)
                return metric_res
            if request.choice in self.protocol_metrics.get_supported_protocols():
                response = self.protocol_metrics.results(request)
                metric_res = self.metrics_response()
                getattr(
                    metric_res, request.choice + '_metrics'
                ).deserialize(response)
                return metric_res
        except Exception as err:
            raise SnappiIxnException(err)
        if request.choice is not None:
            msg = "{} is not a supported choice for metrics; \
            the supported choices are \
            ['port', 'flow']".format(request.choice)
            raise SnappiIxnException(400, msg)

    def add_error(self, error):
        """Add an error to the global errors
        """
        if isinstance(error, str) is False:
            self._errors.append('%s %s' % (type(error), str(error)))
        else:
            self._errors.append(error)

    def parse_location_info(self, location):
        """It will return (chassis,card,port)
        set card as 0 where that is not applicable"""
        if re.search("/|;", location) is None:
            raise Exception("Invalid port location format, expected ["
                    "chassis_ip;card_id;port_id or chassis_ip/port_id]")
        LocationInfo = namedtuple("LocationInfo", ["chassis_info",
                                                   "card_info",
                                                   "port_info"])
        if ';' in location:
            try:
                (chassis_info, card_info, port_info) = location.split(';')
            except Exception:
                raise Exception("Please specify <chassis>;<card>;<port>")
        else:
            try:
                card_info = 0
                (chassis_info, port_info) = location.split('/')
            except Exception:
                raise Exception("Please specify <chassis>/<port>")
        return LocationInfo(chassis_info,
                        card_info,
                        port_info)

    def special_char(self, names):
        is_names = True
        if not isinstance(names, list):
            is_names = False
            names = [names]
        
        ret_list = []
        for name in names:
            if name is None:
                ret_list.append(name)
            else:
                ret_list.append(
                    name.replace('(', '\\(').replace(')', '\\)')
                        .replace('[', '\\[').replace(']', '\\]')
                        .replace('.', '\\.').replace('*', '\\*')
                        .replace('+', '\\+').replace('?', '\\?')
                        .replace('{', '\\{').replace('}', '\\}')
                )
        if is_names is True:
            return ret_list
        else:
            return ret_list[0]
    
    def _connect(self):
        """Connect to an IxNetwork API Server.
        """
        self._errors = []
        if self._assistant is None:
            platform = TestPlatform(self._address, rest_port=self._port)
            platform.Authenticate(self._username, self._password)
            url = '%s://%s:%s/ixnetworkweb/api/v1/usersettings/ixnrest' % \
                (platform.Scheme, platform.Hostname, platform.RestPort)
            platform._connection._session.request(
                'put',
                url,
                data=json.dumps({'enableClassicProtocols': True}),
                verify=False)
            self._assistant = SessionAssistant(IpAddress=self._address,
                                               RestPort=self._port,
                                               UserName=self._username,
                                               Password=self._password,
                                               LogLevel=self._log_level)
            self._ixnetwork = self._assistant.Session.Ixnetwork
            self._vport = self._ixnetwork.Vport
            self._lag = self._ixnetwork.Lag
            self._topology = self._ixnetwork.Topology
            self._traffic = self._ixnetwork.Traffic
            self._traffic_item = self._ixnetwork.Traffic.TrafficItem
            self._globals = self._ixnetwork.Globals
            if not self._ixn_version_check():
                raise Exception("IxNetwork 9.10 or newer is required for snappi[ixnetwork]")
            if len(self._license_servers) > 0:
                self._ixnetwork.Globals.Licensing \
                    .LicensingServers = self._license_servers
            try:
                import pkg_resources
                snappi_ver = "snappi-" +\
                    pkg_resources.get_distribution("snappi").version
                self.info(snappi_ver)
                snappi_ixn = "snappi_ixnetwork-" +\
                    pkg_resources.get_distribution("snappi_ixnetwork").version
                self.info(snappi_ixn)
                restpy = "ixnetwork_restpy-" + pkg_resources.get_distribution(
                    "ixnetwork_restpy"
                ).version
                self.info(restpy)
            except pkg_resources.DistributionNotFound as e:
                version = "Could not determine version for pkg {}".format(
                    e.req.project_name
                )
                self.info(version)
            except Exception as e:
                self.warning('{}'.format(e))
        self._backup_errors()

    def _ixn_version_check(self):
        major, minor = self._globals.BuildNumber.split('.')[0:2]
        if int(major) < 9:
            return False
        if int(major) == 9 and int(minor) < 10:
            return False
        return True
    
    def _backup_errors(self):
        app_errors = self._globals.AppErrors.find()
        if len(app_errors) > 0:
            self._ixn_errors = app_errors[0].Error.find()

    def _validate_instance(self, config):
        """Validate current IxNetwork instance:
        1. Stop everything if local config is None
        2. Otherwise add warning message """
        traffic_state = self._traffic.State
        if self.snappi_config is None:
            if traffic_state == 'started':
                self._traffic_item.find()
                if len(self._traffic_item) > 0:
                    self._traffic_item.StopStatelessTrafficBlocking()
            glob_topo = self._globals.Topology.refresh()
            if glob_topo.Status  == 'started':
                self._ixnetwork.StopAllProtocols('sync')
        else:
            if traffic_state == 'started':
                msg = "Flows are in running state. " \
                      "Please stop those using set_transmit_state"
                self.add_error(msg)
                self.warning(msg)
        return config

    def _apply_change(self):
        """Apply on the fly only applicable for Device object"""
        glob_topo = self._globals.Topology.refresh()
        if glob_topo.ApplyOnTheFlyState == 'allowed':
            url = '%s/globals/topology/operations/applyonthefly' % self._ixnetwork.href
            payload = {'arg1': glob_topo.href}
            # todo: Sometime it redirect to some unknown loaction
            try:
                self._request('POST', url, payload)
            except Exception:
                pass
        
    def _request(self, method, url, payload=None):
        connection, url = self._assistant.Session._connection._normalize_url(
            url)
        headers = {}
        if payload is not None:
            payload = json.dumps(payload)
            headers['Content-Type'] = 'application/json'
        response = self._assistant.Session._connection._session.request(
            method, url, headers=headers, data=payload, verify=False)
        response.raise_for_status()
        if response.status_code == 202:
            content = response.json()
            while content['state'] == 'IN_PROGRESS':
                time.sleep(1)
                response = self._request('GET', content['url'])
        if response.headers.get('Content-Type'):
            if response.headers['Content-Type'] == 'application/json':
                return response.json()
            elif response.headers[
                    'Content-Type'] == 'application/octet-stream':
                return response.content
        return response

    def _remove(self, ixn_obj, items):
        """Remove any ixnetwork objects that are not found in the items list.
        If the items list does not exist remove everything.
        """
        valid_names = [item.name for item in items
                       if item.name is not None]
        invalid_names = []
        for item in ixn_obj.find():
            if item.Name not in valid_names:
                invalid_names.append(item.Name)
        if len(invalid_names) > 0:
            if ixn_obj._SDM_NAME == 'trafficItem':
                # can't remove traffic items that are started
                start_states = [
                    'txStopWatchExpected', 'locked', 'started',
                    'startedWaitingForStats', 'startedWaitingForStreams',
                    'stoppedWaitingForStats'
                ]
                for item in ixn_obj.find(Name='^(%s)$' %
                                         '|'.join(self.special_char(invalid_names))):
                    if item.State in start_states:
                        item.StopStatelessTraffic()
                if len(ixn_obj) > 0:
                    poll = True
                    while poll:
                        poll = False
                        for v in self.select_traffic_items().values():
                            if v['state'] not in [
                                    'error', 'stopped', 'unapplied'
                            ]:
                                poll = True
            ixn_obj.find(Name='^(%s)$' % '|'.join(self.special_char(invalid_names)))
            if len(ixn_obj) > 0:
                ixn_obj.remove()

    def _get_topology_name(self, port_name):
        return 'Topology %s' % port_name

    def select_card_aggregation(self, location):
        (hostname, cardid, portid) = location.split(';')
        payload = {
            'selects': [{
                'from':
                '/availableHardware',
                'properties': [],
                'children': [{
                    'child':
                    'chassis',
                    'properties': [],
                    'filters': [{
                        'property': 'hostname',
                        'regex': '^%s$' % hostname
                    }]
                }, {
                    'child':
                    'card',
                    'properties': ['*'],
                    'filters': [{
                        'property': 'cardId',
                        'regex': '^%s$' % abs(int(cardid))
                    }]
                }, {
                    'child':
                    'aggregation',
                    'properties': ['*'],
                    'filters': []
                }],
                'inlines': []
            }]
        }
        url = '%s/operations/select?xpath=true' % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        return results[0]['chassis'][0]['card'][0]

    def select_chassis_card(self, vport):
        pieces = vport['connectionStatus'].split(';')
        payload = {
            'selects': [{
                'from':
                '/availableHardware',
                'properties': [],
                'children': [{
                    'child':
                    'chassis',
                    'properties': [],
                    'filters': [{
                        'property': 'hostname',
                        'regex': '^%s$' % pieces[0]
                    }]
                }, {
                    'child':
                    'card',
                    'properties': ['*'],
                    'filters': [{
                        'property': 'cardId',
                        'regex': '^%s$' % int(pieces[1])
                    }]
                }],
                'inlines': []
            }]
        }
        url = '%s/operations/select?xpath=true' % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        return results[0]['chassis'][0]['card'][0]

    def select_vports(self, port_name_filters=[]):
        """Select all vports.
        Return them in a dict keyed by vport name.
        """
        payload = {
            'selects': [{
                'from':
                '/',
                'properties': [],
                'children': [{
                    'child':
                    'vport',
                    'properties': [
                        'name', 'type', 'location', 'connectionState',
                        'connectionStatus', 'assignedTo', 'connectedTo'
                    ],
                    'filters': port_name_filters
                }, {
                    'child': 'l1Config',
                    'properties': ['currentType'],
                    'filters': []
                }, {
                    'child':
                    'capture',
                    'properties': ['hardwareEnabled', 'softwareEnabled'],
                    'filters': []
                }, {
                    'child': '^(eth.*|novus.*|uhd.*|atlas.*|ares.*|star.*)$',
                    'properties': ['*'],
                    'filters': []
                }],
                'inlines': []
            }]
        }
        url = '%s/operations/select?xpath=true' % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        vports = {}
        if 'vport' in results[0]:
            for vport in results[0]['vport']:
                vports[vport['name']] = vport
        return vports

    def select_traffic_items(self, traffic_item_filters=[]):
        """Select all traffic items.
        Return them in a dict keyed by traffic item name.

        Args
        ----
        - filters (list(dict(property:'', 'regex':''))): A list of filters for the select.
            A filter is a dict with a property name and a regex match
        """
        payload = {
            'selects': [{
                'from':
                '/traffic',
                'properties': [],
                'children': [{
                    'child': 'trafficItem',
                    'properties': ['name', 'state', 'enabled'],
                    'filters': traffic_item_filters
                }, {
                    'child':
                    'highLevelStream',
                    'properties': ['txPortName', 'rxPortNames', 'state'],
                    'filters': []
                }],
                'inlines': []
            }]
        }
        url = '%s/operations/select?xpath=true' % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        traffic_items = {}
        try:
            for traffic_item in results[0]['trafficItem']:
                traffic_items[traffic_item['name']] = traffic_item
        except Exception:
            pass
        return traffic_items

    def select_chassis_card_port(self, location):
        """Select all availalehardware.
        Return them in a dict keyed by vport name.
        """
        (hostname, cardid, portid) = location.split(';')
        payload = {
            'selects': [{
                'from':
                '/availableHardware',
                'properties': [],
                'children': [{
                    'child':
                    'chassis',
                    'properties': [],
                    'filters': [{
                        'property': 'hostname',
                        'regex': '^%s$' % hostname
                    }]
                }, {
                    'child':
                    'card',
                    'properties': [],
                    'filters': [{
                        'property': 'cardId',
                        'regex': '^%s$' % abs(int(cardid))
                    }]
                }, {
                    'child':
                    'port',
                    'properties': [],
                    'filters': [{
                        'property': 'portId',
                        'regex': '^%s$' % abs(int(portid))
                    }]
                }],
                'inlines': []
            }]
        }
        url = '%s/operations/select?xpath=true' % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        return results[0]['chassis'][0]['card'][0]['port'][0]['xpath']

    def clear_ownership(self, available_hardware_hrefs, location_hrefs):
        hrefs = list(available_hardware_hrefs.values()) + list(
            location_hrefs.values())
        if len(hrefs) == 0:
            return
        names = list(available_hardware_hrefs.keys()) + list(
            location_hrefs.keys())
        with Timer(self, 'Location preemption [%s]' % ', '.join(names)):
            payload = {'arg1': [href for href in hrefs]}
            url = '%s/operations/clearownership' % payload['arg1'][0]
            self._ixnetwork._connection._execute(url, payload)

    def get_config(self):
        return self._config

    def check_protocol_statistics(self):
        start = time.time()
        url = '%s/operations/gettopologystatus' % self._ixnetwork.href
        check = True
        while check is True and time.time() - start < 90:
            check = False
            results = self._ixnetwork._connection._execute(url, None)
            for result in results:
                if result['arg2'][0]['arg2'] != result['arg2'][3]['arg2']:
                    check = True

    def info(self, message):
        self._ixnetwork.info('[ixn-otg] %s' % message)

    def warning(self, message):
        self._ixnetwork.warn('[ixn-otg] %s' % message)



import json
import time
import re
from snappi_ixnetwork.timer import Timer


class Vport(object):
    """Transforms OpenAPI objects into IxNetwork objects

    Args
    ----
    - ixnetworkapi (Api): instance of the Api class

    Transformations
    ---------------
    - /components/schemas/Port to /vport
    - /components/schemas/Layer1 to /vport/l1Config/...

    Process
    -------
    - Remove any vports that are not in the config.ports
    - Add any vports that are in the config.ports
    - If the location of the config.ports.location is different than the
      the /vport -connectedTo property set it to None
    - If the config.ports.location is None don't connect the ports
      else connect the port, get the vport type, set the card mode based on the
      config.layer1.speed

    Notes
    -----
    - Uses resourcemanager to set the vport location and l1Config as it is the
      most efficient way. DO NOT use the AssignPorts API as it is too slow. 
    - Only setup l1Config if location is connected. 
    - Given a connected location and speed the vport -type, card resource mode
      and l1Config sub node are derived.

    """
    _SPEED_MAP = {
        'speed_400_gbps': 'speed400g',
        'speed_200_gbps': 'speed200g',
        'speed_100_gbps': 'speed100g',
        'speed_50_gbps': 'speed50g',
        'speed_40_gbps': 'speed40g',
        'speed_25_gbps': 'speed25g',
        'speed_10_gbps': 'speed10g',
        'speed_1_gbps': 'speed1000',
        'speed_100_fd_mbps': 'speed100fd',
        'speed_100_hd_mbps': 'speed100hd',
        'speed_10_fd_mbps': 'speed10fd',
        'speed_10_hd_mbps': 'speed10hd'
    }
    _VM_SPEED_MAP = {
        'speed_400_gbps': 'speed400g',
        'speed_200_gbps': 'speed200g',
        'speed_100_gbps': 'speed100g',
        'speed_90_gbps': 'speed90g',
        'speed_80_gbps': 'speed80g',
        'speed_70_gbps': 'speed70g',
        'speed_60_gbps': 'speed60g',
        'speed_50_gbps': 'speed50g',
        'speed_40_gbps': 'speed40g',
        'speed_30_gbps': 'speed30g',
        'speed_25_gbps': 'speed25g',
        'speed_20_gbps': 'speed20g',
        'speed_10_gbps': 'speed10g',
        'speed_9_gbps': 'speed9000',
        'speed_8_gbps': 'speed8000',
        'speed_7_gbps': 'speed7000',
        'speed_6_gbps': 'speed6000',
        'speed_5_gbps': 'speed5000',
        'speed_4_gbps': 'speed4000',
        'speed_3_gbps': 'speed3000',
        'speed_2_gbps': 'speed2000',
        'speed_1_gbps': 'speed1000',
        'speed_100_mbps': 'speed100',
        'speed_100_fd_mbps': 'speed100',
        'speed_100_hd_mbps': 'speed100',
        'speed_10_fd_mbps': 'speed100',
        'speed_10_hd_mbps': 'speed100'
    }

    _SPEED_MODE_MAP = {
        'speed_1_gbps': 'normal',
        'speed_10_gbps': 'tengig',
        'speed_25_gbps': 'twentyfivegig',
        'speed_40_gbps': 'fortygig',
        'speed_50_gbps': 'fiftygig',
        'speed_100_gbps':
            '^(?!.*(twohundredgig|fourhundredgig)).*hundredgig.*$',
        'speed_200_gbps': 'twohundredgig',
        'speed_400_gbps': 'fourhundredgig'
    }
    
    _ADVERTISE_MAP = {
        'advertise_one_thousand_mbps': 'speed1000',
        'advertise_one_hundred_fd_mbps': 'speed100fd',
        'advertise_one_hundred_hd_mbps': 'speed100hd',
        'advertise_ten_fd_mbps': 'speed10fd',
        'advertise_ten_hd_mbps': 'speed10hd'
    }
    _FLOW_CONTROL_MAP = {
        'ieee_802_1qbb': 'ieee802.1Qbb',
        'ieee_802_3x': 'ieee802.3x'
    }

    _RESULT_COLUMNS = [
        ('frames_tx', 'Frames Tx.', int),
        ('frames_rx', 'Valid Frames Rx.', int),
        ('frames_tx_rate', 'Frames Tx. Rate', float),
        ('frames_rx_rate', 'Valid Frames Rx. Rate', float),
        ('bytes_tx', 'Bytes Tx.', int),
        ('bytes_rx', 'Bytes Rx.', int),
        ('bytes_tx_rate', 'Bytes Tx. Rate', float),
        ('bytes_rx_rate', 'Bytes Rx. Rate', float),
        ('pfc_class_0_frames_rx', 'Rx Pause Priority Group 0 Frames', int),
        ('pfc_class_1_frames_rx', 'Rx Pause Priority Group 1 Frames', int),
        ('pfc_class_2_frames_rx', 'Rx Pause Priority Group 2 Frames', int),
        ('pfc_class_3_frames_rx', 'Rx Pause Priority Group 3 Frames', int),
        ('pfc_class_4_frames_rx', 'Rx Pause Priority Group 4 Frames', int),
        ('pfc_class_5_frames_rx', 'Rx Pause Priority Group 5 Frames', int),
        ('pfc_class_6_frames_rx', 'Rx Pause Priority Group 6 Frames', int),
        ('pfc_class_7_frames_rx', 'Rx Pause Priority Group 7 Frames', int),
    ]

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self._layer1_check = []

    def config(self):
        """Transform config.ports into Ixnetwork.Vport
        1) delete any vport that is not part of the config
        2) create a vport for every config.ports[] not present in IxNetwork
        3) set config.ports[].location to /vport -location or -connectedTo
        4) set /vport/l1Config/... properties using the corrected /vport -type
        5) connectPorts to use new l1Config settings and clearownership
        """
        self._resource_manager = self._api._ixnetwork.ResourceManager
        self._ixn_vport = self._api._vport
        self._layer1_check = []
        with Timer(self._api, 'Ports configuration'):
            self._delete_vports()
            self._create_vports()
        with Timer(self._api, 'Captures configuration'):
            self._api.capture.config()
        with Timer(self._api, 'Location configuration'):
            self._set_location()
        with Timer(self._api, 'Layer1 configuration'):
            self._set_layer1()

    def set_link_state(self, link_state):
        with Timer(self._api, 'Link State operation'):
            payload = {
                'arg1': [],
                'arg2': link_state.state,
            }
            for port_name in link_state.port_names:
                payload['arg1'].append(self._api.get_ixn_href(port_name))
            url = '%s/vport/operations/linkupdn' % self._api._ixnetwork.href
            self._api._request('POST', url, payload)
    
    def _import(self, imports):
        if len(imports) > 0:
            errata = self._resource_manager.ImportConfig(
                json.dumps(imports), False)
            for item in errata:
                self._api.warning(item)
            return len(errata) == 0
        return True

    def _delete_vports(self):
        """Delete any vports from the api server that do not exist in the new config
        """
        self._api._remove(self._ixn_vport, self._api.snappi_config.ports)

    def _create_vports(self):
        """Add any vports to the api server that do not already exist
        """
        vports = self._api.select_vports()
        imports = []
        for port in self._api.snappi_config.ports:
            if port.name not in vports.keys():
                index = len(vports) + len(imports) + 1
                vport_import = {
                    'xpath': '/vport[%i]' % index,
                    'name': port.name,
                    'rxMode': 'captureAndMeasure',
                    'txMode': 'interleaved'
                }
                location = getattr(port, 'location', None)
                if location is None:
                    vport_import['connectedTo'] = location
                    port.location = None
                imports.append(vport_import)
        self._import(imports)
        for name, vport in self._api.select_vports().items():
            self._api.ixn_objects[name] = vport['href']

    def _add_hosts(self, HostReadyTimeout):
        chassis = self._api._ixnetwork.AvailableHardware.Chassis
        add_addresses = []
        check_addresses = []
        for port in self._api.snappi_config.ports:
            location = getattr(port, 'location', None)
            if location is not None:
                location_info = self._api.parse_location_info(
                        location)
                chassis_address = location_info.chassis_info
                chassis.find(Hostname='^%s$' % chassis_address)
                if len(chassis) == 0:
                    add_addresses.append(chassis_address)
                check_addresses.append(chassis_address)
        add_addresses = set(add_addresses)
        check_addresses = set(check_addresses)
        if len(add_addresses) > 0:
            with Timer(self._api,
                       'Add location hosts [%s]' % ', '.join(add_addresses)):
                for add_address in add_addresses:
                    chassis.add(Hostname=add_address)
        if len(check_addresses) > 0:
            with Timer(
                    self._api,
                    'Location hosts ready [%s]' % ', '.join(check_addresses)):
                start_time = time.time()
                while True:
                    chassis.find(Hostname='^(%s)$' % '|'.join(check_addresses),
                                 State='^ready$')
                    if len(chassis) == len(check_addresses):
                        break
                    if time.time() - start_time > HostReadyTimeout:
                        raise RuntimeError(
                            'After %s seconds, not all location hosts [%s] are reachable'
                            % (HostReadyTimeout, ', '.join(check_addresses)))
                    time.sleep(2)
    
    def _set_location(self):
        location_supported = True
        try:
            self._api._ixnetwork._connection._options(
                self._api._ixnetwork.href + '/locations')
        except Exception:
            location_supported = False
        
        self._add_hosts(60)
        with Timer(self._api,
                   'Aggregation mode speed change'):
            layer1_check = self._api.resource_group.set_group()
            self._layer1_check.extend(layer1_check)
        vports = self._api.select_vports()
        locations = []
        imports = []
        clear_locations = []
        for port in self._api.snappi_config.ports:
            vport = vports[port.name]
            location = getattr(port, 'location', None)

            if location_supported is True:
                if vport['location'] == location and vport[
                        'connectionState'].startswith('connectedLink'):
                    continue
            else:
                if len(vport['connectedTo']) > 0 and vport[
                        'connectionState'].startswith('connectedLink'):
                    continue
                    
            self._api.ixn_objects[port.name] = vport['href']
            vport = {'xpath': vports[port.name]['xpath']}
            if location_supported is True:
                vport['location'] = location
            else:
                if location is not None:
                    xpath = self._api.select_chassis_card_port(location)
                    vport['connectedTo'] = xpath
                else:
                    vport['connectedTo'] = ''
            imports.append(vport)
            if location is not None and len(location) > 0:
                clear_locations.append(location)
                locations.append(port.name)
        if len(locations) == 0:
            return
        self._clear_ownership(clear_locations)
        with Timer(self._api, 'Location connect [%s]' % ', '.join(locations)):
            self._import(imports)
        with Timer(self._api,
                   'Location state check [%s]' % ', '.join(locations)):
            self._api._vport.find(ConnectionState='^(?!connectedLink).*$')
            if len(self._api._vport) > 0:
                self._api._vport.ConnectPorts()
            start = time.time()
            timeout = 10
            while True:
                self._api._vport.find(Name='^(%s)$' % '|'.join(self._api.special_char(locations)),
                                      ConnectionState='^connectedLink')
                if len(self._api._vport) == len(locations):
                    break
                if time.time() - start > timeout:
                    unreachable = []
                    self._api._vport.find(ConnectionState='^(?!connectedLink).*$')
                    for vport in self._api._vport:
                        unreachable.append('%s [%s: %s]' % (vport.Name, vport.ConnectionState, vport.ConnectionStatus))
                    raise RuntimeError(
                        'After %s seconds, %s are unreachable' % (timeout, ', '.join(unreachable)))
                time.sleep(2)
            for vport in self._api._vport.find(
                    ConnectionState='^(?!connectedLinkUp).*$'):
                self._api.warning('%s %s' %
                                  (vport.Name, vport.ConnectionState))

    def _set_layer1(self):
        """Set the /vport/l1Config/... properties
        This should only happen if the vport connectionState is connectedLink...
        as it determines the ./l1Config child node.
        """
        if len(self._api.snappi_config.layer1) == 0:
            return
        reset_auto_negotiation = dict()
        # set and commit the card resource mode
        vports = self._api.select_vports()
        imports = []
        for layer1 in self._api.snappi_config.layer1:
            for port_name in layer1.port_names:
                self._set_card_resource_mode(vports[port_name], layer1,
                                             imports)
        if self._import(imports) is False:
            # WARNING: this retry is because no reasonable answer as to why
            # changing card mode periodically fails with this opaque message
            # 'Releasing ownership on ports failed.'
            self._api.info('Retrying card resource mode change')
            self._import(imports)
        # set the vport type
        imports = []
        for layer1 in self._api.snappi_config.layer1:
            for port_name in layer1.port_names:
                self._set_vport_type(vports[port_name], layer1, imports)
        self._import(imports)
        vports = self._api.select_vports()
        # set the remainder of l1config properties
        imports = []
        for layer1 in self._api.snappi_config.layer1:
            for port_name in layer1.port_names:
                self._set_l1config_properties(vports[port_name], layer1,
                                              imports)
        self._import(imports)
        # Due to dependency attribute (ieeeL1Defaults)
        # reset enableAutoNegotiation
        imports = []
        for layer1 in self._api.snappi_config.layer1:
            for port_name in layer1.port_names:
                vport = vports[port_name]
                if port_name in reset_auto_negotiation and reset_auto_negotiation[
                        port_name]:
                    self._reset_auto_negotiation(vport, layer1, imports)
        self._import(imports)

    def _set_l1config_properties(self, vport, layer1, imports):
        """Set vport l1config properties
        """
        if vport['connectionState'] not in [
                'connectedLinkUp', 'connectedLinkDown'
        ]:
            return
        self._set_fcoe(vport, layer1, imports)
        self._import(imports)
        
        self._set_auto_negotiation(vport, layer1, imports)
    
    def _set_card_resource_mode(self, vport, layer1, imports):
        """If the card has an aggregation mode set it according to the speed
        """
        if vport['connectionState'] not in [
                'connectedLinkUp', 'connectedLinkDown'
        ] or layer1.name in self._layer1_check:
            return

        aggregation_mode = None
        if layer1.speed in Vport._SPEED_MODE_MAP:
            card = self._api.select_chassis_card(vport)
            mode = Vport._SPEED_MODE_MAP[layer1.speed]
            for available_mode in card['availableModes']:
                if re.search(mode, available_mode.lower()) is not None:
                    aggregation_mode = available_mode
                    break
        if aggregation_mode is not None and aggregation_mode != card[
                'aggregationMode']:
            self._api.info('Setting %s to resource mode %s' %
                           (card['description'], aggregation_mode))
            imports.append({
                'xpath': card['xpath'],
                'aggregationMode': aggregation_mode
            })

    def _set_auto_negotiation(self, vport, layer1, imports):
        if layer1.speed.endswith('_mbps') or layer1.speed == 'speed_1_gbps':
            self._set_ethernet_auto_negotiation(vport, layer1, imports)
        else:
            self._set_gigabit_auto_negotiation(vport, layer1, imports)

    def _set_vport_type(self, vport, layer1, imports):
        """Set the /vport -type
        
        If flow_control is not None then the -type attribute should 
        be switched to a type with the Fcoe extension if it is allowed.

        If flow_control is None then the -type attribute should 
        be switched to a type without the Fcoe extension.
        """
        fcoe = False
        if layer1.flow_control.choice is not None:
            fcoe = True
        vport_type = vport['type']
        elegible_fcoe_vport_types = [
            'ethernet', 'tenGigLan', 'fortyGigLan', 'tenGigWan',
            'hundredGigLan', 'tenFortyHundredGigLan', 'novusHundredGigLan',
            'novusTenGigLan', 'krakenFourHundredGigLan',
            'aresOneFourHundredGigLan', 'starFourHundredGigLan'
        ]
        if fcoe is True and vport_type in elegible_fcoe_vport_types:
            vport_type = vport_type + 'Fcoe'
        if fcoe is False and vport_type.endswith('Fcoe'):
            vport_type = vport_type.replace('Fcoe', '')
        if vport_type != vport['type']:
            imports.append({
                'xpath': vport['xpath'] + '/l1Config',
                'currentType': vport_type
            })
        return vport_type

    def _set_ethernet_auto_negotiation(self, vport, layer1, imports):
        advertise = []
        if layer1.speed == 'speed_1_gbps':
            advertise.append(
                Vport._ADVERTISE_MAP['advertise_one_thousand_mbps'])
        if layer1.speed == 'speed_100_fd_mbps':
            advertise.append(
                Vport._ADVERTISE_MAP['advertise_one_hundred_fd_mbps'])
        if layer1.speed == 'speed_100_hd_mbps':
            advertise.append(
                Vport._ADVERTISE_MAP['advertise_one_hundred_hd_mbps'])
        if layer1.speed == 'speed_10_fd_mbps':
            advertise.append(Vport._ADVERTISE_MAP['advertise_ten_fd_mbps'])
        if layer1.speed == 'speed_10_hd_mbps':
            advertise.append(Vport._ADVERTISE_MAP['advertise_ten_hd_mbps'])
        proposed_import = {
            'xpath':
            vport['xpath'] + '/l1Config/' + vport['type'].replace('Fcoe', ''),
            'speed':
            self._get_speed(vport, layer1),
            'media':
            layer1.media,
            'autoNegotiate':
            layer1.auto_negotiate,
            'speedAuto':
            advertise
        }
        self._add_l1config_import(vport, proposed_import, imports)

    def _add_l1config_import(self, vport, proposed_import, imports):
        type = vport['type'].replace('Fcoe', '')
        l1config = vport['l1Config'][type]
        key_to_remove = []
        for key in proposed_import:
            if key == 'xpath':
                continue
            if key not in l1config or l1config[key] == proposed_import[key]:
                key_to_remove.append(key)
        # add this constrain due to handle some specific use case (1G to 10G)
        if 'speed' in key_to_remove and 'speedAuto' not in key_to_remove:
            key_to_remove.remove('speed')
        for key in key_to_remove:
            proposed_import.pop(key)
        if len(proposed_import) > 0:
            imports.append(proposed_import)

    def _set_gigabit_auto_negotiation(self, vport, layer1, imports):
        advertise = []
        advertise.append(Vport._SPEED_MAP[layer1.speed])
        auto_field_name  = 'enableAutoNegotiation'
        if re.search('novustengiglan', vport['type'].lower()) is not None:
            auto_field_name = 'autoNegotiate'
        # Due to ieeeL1Defaults dependency
        ieee_media_defaults = {
            'xpath':
                vport['xpath'] + '/l1Config/' + vport['type'].replace('Fcoe', ''),
            'ieeeL1Defaults':
                layer1.ieee_media_defaults
        }
        self._add_l1config_import(vport, ieee_media_defaults, imports)
        proposed_import = {
            'xpath':
            vport['xpath'] + '/l1Config/' + vport['type'].replace('Fcoe', ''),
            'speed':
            Vport._SPEED_MAP[layer1.speed],
            '{0}'.format(auto_field_name):
            False if layer1.auto_negotiate is None else layer1.auto_negotiate,
            'enableRsFec':
            False if layer1.auto_negotiation.rs_fec is None else
            layer1.auto_negotiation.rs_fec,
            'linkTraining':
            False if layer1.auto_negotiation.link_training is None else
            layer1.auto_negotiation.link_training,
            'speedAuto':
            advertise
        }
        if layer1.media is not None:
            proposed_import['media'] = layer1.media
            
        self._add_l1config_import(vport, proposed_import, imports)

    def _get_speed(self, vport, layer1):
        if vport['type'] == 'ethernetvm':
            return Vport._VM_SPEED_MAP[layer1.speed]
        else:
            return Vport._SPEED_MAP[layer1.speed]

    def _reset_auto_negotiation(self, vport, layer1, imports):
        if layer1.speed.endswith(
                '_mbps') is False and layer1.speed != 'speed_1_gbps':
            imports.append({
                'xpath':
                vport['xpath'] + '/l1Config/' +
                vport['type'].replace('Fcoe', ''),
                'enableAutoNegotiation':
                layer1.auto_negotiate,
            })

    def _set_fcoe(self, vport, layer1, imports):
        if layer1.flow_control.choice is None:
            return
        xpath = '%s/l1Config/%s/fcoe' % (vport['xpath'], vport['type'].replace(
            'Fcoe', ''))
        fcoe = {
            'xpath': xpath,
            'flowControlType':
            Vport._FLOW_CONTROL_MAP[layer1.flow_control.choice]
        }
        if layer1.flow_control.choice == 'ieee_802_1qbb':
            pfc = layer1.flow_control.ieee_802_1qbb
            fcoe['enablePFCPauseDelay'] = False if pfc.pfc_delay == 0 else True
            fcoe['pfcPauseDelay'] = pfc.pfc_delay
            fcoe['pfcPriorityGroups'] = [
                -1 if pfc.pfc_class_0 is None else pfc.pfc_class_0,
                -1 if pfc.pfc_class_1 is None else pfc.pfc_class_1,
                -1 if pfc.pfc_class_2 is None else pfc.pfc_class_2,
                -1 if pfc.pfc_class_3 is None else pfc.pfc_class_3,
                -1 if pfc.pfc_class_4 is None else pfc.pfc_class_4,
                -1 if pfc.pfc_class_5 is None else pfc.pfc_class_5,
                -1 if pfc.pfc_class_6 is None else pfc.pfc_class_6,
                -1 if pfc.pfc_class_7 is None else pfc.pfc_class_7,
            ]
            fcoe['priorityGroupSize'] = 'priorityGroupSize-8'
            fcoe['supportDataCenterMode'] = True
        imports.append(fcoe)

    def _clear_ownership(self, locations):
        try:
            force_ownership = self._api.snappi_config.options.port_options.location_preemption
        except Exception:
            force_ownership = False
        if force_ownership is True:
            available_hardware_hrefs = {}
            location_hrefs = {}
            for location in locations:
                if ';' in location:
                    clp = location.split(';')
                    chassis = self._api._ixnetwork.AvailableHardware.Chassis.find(
                        Hostname=clp[0])
                    if len(chassis) > 0:
                        available_hardware_hrefs[
                            location] = '%s/card/%s/port/%s' % (
                                chassis.href, abs(int(clp[1])), abs(int(
                                    clp[2])))
                elif '/' in location:
                    appliance = location.split('/')[0]
                    locations = self._api._ixnetwork.Locations
                    locations.find(Hostname=appliance)
                    if len(locations) == 0:
                        locations.add(Hostname=appliance)
                    ports = locations.Ports.find(Location='^%s$' % location)
                    if len(ports) > 0:
                        location_hrefs[location] = ports.href
            self._api.clear_ownership(available_hardware_hrefs, location_hrefs)

    def _set_result_value(self,
                          row,
                          column_name,
                          column_value,
                          column_type=str):
        if len(self._column_names
               ) > 0 and column_name not in self._column_names:
            return
        try:
            row[column_name] = column_type(column_value)
        except Exception:
            if column_type.__name__ in ['float', 'int']:
                row[column_name] = 0
            else:
                row[column_type] = column_value

    def results(self, request):
        """Return port results
        """

        self._column_names = request._properties.get('column_names')
        if self._column_names is None:
            self._column_names = []
        elif not isinstance(self._column_names, list):
            msg = "Invalid format of port_names passed {},\
                    expected list".format(self._column_names)
            raise Exception(msg)

        port_names = request._properties.get('port_names')
        if port_names is None or len(port_names) == 0:
            port_names = [port.name for port in self._api._config.ports]
        elif not isinstance(port_names, list):
            msg = "Invalid format of port_names passed {},\
                    expected list".format(port_names)
            raise Exception(msg)

        port_filter = {'property': 'name', 'regex': '.*'}
        port_filter['regex'] = '^(%s)$' % '|'.join(self._api.special_char(
                            port_names))

        port_rows = dict()
        vports = self._api.select_vports(port_name_filters=[port_filter])
        for vport in vports.values():
            port_row = dict()
            self._set_result_value(port_row, 'name', vport.get('name'))
            location = vport.get('location')
            if vport.get(
                'connectionState'
            ).startswith('connectedLink') is True:
                location += ';connected'
            elif len(location) > 0:
                location += ';' + vport.get('connectionState')
            else:
                location = vport.get('connectionState')
            self._set_result_value(port_row, 'location', location)
            self._set_result_value(
                port_row, 'link', 'up'
                if vport['connectionState'] == 'connectedLinkUp' else 'down')
            self._set_result_value(port_row, 'capture', 'stopped')
            # init all columns with corresponding zero-values so that
            # the underlying dictionary contains all requested columns
            # in an event of unwanted exceptions
            for ext_name, _, typ in self._RESULT_COLUMNS:
                self._set_result_value(port_row, ext_name, 0, typ)

            port_rows[vport['name']] = port_row

        try:
            table = self._api.assistant.StatViewAssistant('Port Statistics')
        except Exception:
            self._api.warning("Could not retrive the port statistics viewer")
            return list(port_rows.values())

        for row in table.Rows:
            vport_name = row['Port Name']
            if vport_name is None:
                raise Exception("Could not retrive 'Port Name' from stats")
            port_row = port_rows.get(vport_name)
            if port_row is None:
                continue
            for ext_name, int_name, typ in self._RESULT_COLUMNS:
                try:
                    row_val = row[int_name]
                    self._set_result_value(port_row, ext_name, row_val, typ)
                except Exception:
                    # TODO print a warning maybe ?
                    pass
        return list(port_rows.values())
